#############################################################################
# Author: degon & costar
# Created on August 21, 2020, 11:42 AM
# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os
from contextlib import contextmanager

from karabo.common.services import KARABO_CONFIG_MANAGER as MANAGER
from karabo.middlelayer import (
    DaqPolicy, Device, Double, Hash, HashList, KaraboError, Slot, String, call,
    connectDevice, getConfigurationFromName, getLastConfiguration,
    instantiateFromName, listConfigurationFromName,
    listDevicesWithConfiguration, saveConfigurationFromName, slot)
from karabo.middlelayer.tests.eventloop import DeviceTest, async_tst

from ..configuration_manager import ConfigurationManager

DB_NAME = "test_karabo_db"

conf = {
    "classId": "ConfigurationManager",
    "_deviceId_": MANAGER,
    "dbName": DB_NAME
}

conf_alice = {
    "_deviceId_": "ALICE"
}

conf_bob = {
    "_deviceId_": "BOB"
}

conf_charlie = {
    "_deviceId_": "CHARLIE"
}


class TestDevice(Device):
    value = Double(
        defaultValue=5.0,
        daqPolicy=DaqPolicy.SAVE,
        warnHigh=100.0)


DEFAULT_LAST_DOUBLE = -1.0


class MockServer(Device):
    lastClassId = String(
        defaultValue=""
    )

    lastDeviceId = String(
        defaultValue=""
    )

    lastConfigDouble = Double(
        defaultValue=DEFAULT_LAST_DOUBLE
    )

    @Slot()
    async def reset(self):
        self.lastClassId = ""
        self.lastConfigDouble = DEFAULT_LAST_DOUBLE
        self.lastDeviceId = ""

    @slot
    async def slotStartDevice(self, info):
        """The info hash should contain

        - configuration
        - deviceId
        - classId
        """
        deviceId = info["deviceId"]
        config = info["configuration"]
        classId = info["classId"]

        self.lastDeviceId = deviceId
        self.lastClassId = classId
        self.lastConfigDouble = config.get('value', DEFAULT_LAST_DOUBLE)

        return True, self.deviceId

    @slot
    def slotGetClassSchema(self, classId):
        """Fake the retrieval of a class schema retrieval"""
        schema = TestDevice.getClassSchema()
        return schema, "TestDevice", self.deviceId


class TestConfigurationManager(DeviceTest):

    @classmethod
    def tearDownClass(cls):
        db = cls.dev.db.path
        os.remove(db)
        super(cls, cls).tearDownClass()

    @classmethod
    @contextmanager
    def lifetimeManager(cls):
        cls.dev = ConfigurationManager(conf)
        cls.testDev = TestDevice(conf_alice)
        cls.anotherTestDev = TestDevice(conf_bob)
        cls.clientDev = TestDevice(conf_charlie)
        with cls.deviceManager(cls.testDev, cls.clientDev, cls.anotherTestDev,
                               lead=cls.dev):
            yield

    @async_tst
    async def test_configuration_save(self):
        """Test the manual saving of configurations"""
        config_name = "testConfig"

        attrs = [Hash("path", "value", "attribute", "warnHigh", "value", 50.0)]
        r = await call("ALICE", "slotUpdateSchemaAttributes", attrs)
        self.assertEqual(self.testDev.value.descriptor.warnHigh, 50.0)
        h = Hash("name", config_name, "deviceIds", ["ALICE"],
                 "priority", 3)
        r = await call(MANAGER, "slotSaveConfigurationFromName", h)
        self.assertEqual(r["success"], True)
        config_name = "testConfig1"
        h = Hash("name", config_name, "deviceIds", ["ALICE",
                                                    "BOB"],
                 "priority", 2)
        r = await call(MANAGER, "slotSaveConfigurationFromName", h)
        self.assertEqual(r["success"], True)

        invalid_config_name = "s#invalid"
        h = Hash("name", invalid_config_name, "deviceIds", ["ALICE",
                                                            "BOB"],
                 "priority", 2)
        with self.assertRaises(KaraboError):
            await call(MANAGER, "slotSaveConfigurationFromName", h)

        long_name = "s" * 40
        h = Hash("name", long_name, "deviceIds", ["ALICE",
                                                  "BOB"],
                 "priority", 2)
        with self.assertRaises(KaraboError):
            await call(MANAGER, "slotSaveConfigurationFromName", h)

    @async_tst
    async def test_configuration_save_check(self):
        """Cross check if we saved the config"""
        config_name = "testConfig1"
        h = Hash("name", config_name, "deviceIds", ["ALICE",
                                                    "BOB"])
        r = await call(MANAGER, "slotCheckConfigurationFromName", h)
        self.assertEqual(r["success"], True)
        self.assertEqual(r["taken"], True)

        # Wrong config name
        config_name = "NotSavedConfig"
        h = Hash("name", config_name, "deviceIds", ["ALICE",
                                                    "BOB"])
        r = await call(MANAGER, "slotCheckConfigurationFromName", h)
        self.assertEqual(r["success"], True)
        self.assertEqual(r["taken"], False)

        # Wrong device names
        config_name = "testConfig1"
        h = Hash("name", config_name, "deviceIds", ["FOODEVICE",
                                                    "BARDEVICE"])
        r = await call(MANAGER, "slotCheckConfigurationFromName", h)
        self.assertEqual(r["success"], True)
        self.assertEqual(r["taken"], False)

        h = Hash("priority", 3)
        r = await call(MANAGER, "slotListDevices", h)
        self.assertEqual(r["success"], True)
        devices = r["item"]
        self.assertNotIn("FOODEVICE", devices)
        self.assertIn("ALICE", devices)
        self.assertNotIn("BOB", devices)

        h = Hash("priority", 2)
        r = await call(MANAGER, "slotListDevices", h)
        self.assertEqual(r["success"], True)
        devices = r["item"]
        self.assertNotIn("FOODEVICE", devices)
        self.assertIn("ALICE", devices)
        self.assertIn("BOB", devices)

    @async_tst
    async def test_configuration_bulk_save_reject(self):
        """Test the reject of saving of too many configurations"""
        deviceIds = ["ALICE"] * 31
        config_name = "testConfig"
        h = Hash("name", config_name, "deviceIds", deviceIds,
                 "priority", 3)
        with self.assertRaises(KaraboError):
            await call(MANAGER, "slotSaveConfigurationFromName", h)

    @async_tst
    async def test_get_configuration(self):
        """Test the manual retrieving of configurations"""
        config_name = "testConfig"
        h = Hash("name", config_name, "deviceId", "ALICE", "schema",
                 True)
        r = await call(MANAGER, "slotGetConfigurationFromName", h)
        item = r["item"]
        priority = item["priority"]
        self.assertEqual(priority, 3)
        name = item["name"]
        self.assertEqual(name, config_name)
        config = item["config"]
        value = config["value"]
        self.assertEqual(value, 5.0)
        schema = item["schema"]
        policy = schema.hash["value", "daqPolicy"]
        self.assertEqual(policy, DaqPolicy.SAVE)

    @async_tst
    async def test_list_configuration(self):
        h = Hash("name", "", "deviceId", "ALICE")
        r = await call(MANAGER, "slotListConfigurationFromName", h)
        items = r["items"]
        self.assertIsInstance(items, HashList)
        # We stored two configurations!
        self.assertEqual(len(items), 2)
        item = items[0]
        self.assertEqual(item["name"], "testConfig")
        item = items[1]
        self.assertEqual(item["name"], "testConfig1")

        h = Hash("name", "", "deviceId", "BOB")
        r = await call(MANAGER, "slotListConfigurationFromName", h)
        items = r["items"]
        self.assertIsInstance(items, HashList)
        # We stored one configuration!
        self.assertEqual(len(items), 1)

    @async_tst
    async def test_list_configuration_sets(self):
        devices = ["ALICE", "BOB"]
        h = Hash("deviceIds", devices)
        r = await call(MANAGER, "slotListConfigurationSets", h)
        items = r["items"]
        # Just one succesful save for both under same setId
        self.assertEqual(len(items), 1)
        config_set = items[0]
        self.assertEqual(config_set["name"], "testConfig1")
        self.assertEqual(config_set["diff_timepoint"], 0.0)
        self.assertEqual(config_set["description"], "")
        self.assertEqual(config_set["user"], ".")

    @async_tst
    async def test_zhe_get_last_configuration(self):
        """The typo is intended as order of tests is alphabetically done"""
        h = Hash("deviceId", "ALICE")
        r = await call(MANAGER,
                       "slotListConfigurationFromName", h)
        items = r["items"]
        # We stored two configurations!
        self.assertEqual(len(items), 2)

        # Get the last configuration!
        h = Hash("deviceId", "ALICE", "priority", 3, "schema", True)
        r = await call(MANAGER, "slotGetLastConfiguration", h)

        item = r["item"]
        name = item["name"]
        self.assertEqual(name, "testConfig")
        config = item["config"]
        value = config["value"]
        self.assertEqual(value, 5.0)
        schema = item["schema"]
        policy = schema.hash["value", "daqPolicy"]
        self.assertEqual(policy, DaqPolicy.SAVE)

        # Get the last config without schema!
        h = Hash("deviceId", "ALICE", "priority", 3)
        r = await call(MANAGER, "slotGetLastConfiguration", h)
        item = r["item"]
        self.assertNotIn("schema", item)

    @async_tst
    async def test_get_client_configuration(self):
        """Test the functions that can be used from ikarabo"""
        # Device client functions
        await saveConfigurationFromName(
            "CHARLIE", name="testConfigClient",
            description="No desc", priority=3)
        await saveConfigurationFromName(
            ["CHARLIE"], name="testConfigClientList",
            description="Client list", priority=1)

        dev = await connectDevice("CHARLIE")
        await saveConfigurationFromName(
            ["CHARLIE", dev], name="testConfigClientMixedList",
            description="Mixed", priority=2)
        li = await listConfigurationFromName("CHARLIE")
        self.assertEqual(len(li), 3)
        self.assertIsInstance(li, HashList)
        clist = ["testConfigClient", "testConfigClientList",
                 "testConfigClientMixedList"]
        self.assertIn(li[0]["name"], clist)
        self.assertIn(li[1]["name"], clist)
        config = await getConfigurationFromName("CHARLIE",
                                                "testConfigClientList")
        value = config["value"]
        self.assertEqual(value, 5.0)

        item = await getLastConfiguration("CHARLIE", priority=3)
        self.assertEqual(item["name"], "testConfigClient")
        config = item["config"]
        value = config["value"]
        self.assertEqual(value, 5.0)
        self.assertEqual(item["priority"], 3)
        self.assertEqual(item["user"], ".")
        self.assertEqual(item["description"], "No desc")

        deviceIds = await listDevicesWithConfiguration()
        self.assertIn("CHARLIE", deviceIds)
        self.assertIn("ALICE", deviceIds)
        self.assertNotIn("BOB", deviceIds)

        deviceIds = await listDevicesWithConfiguration(priority=2)
        self.assertIn("CHARLIE", deviceIds)
        self.assertIn("ALICE", deviceIds)
        self.assertIn("BOB", deviceIds)

        # Instantiate a device with a device client function
        serverMock = MockServer({"_deviceId_": "MDL_SERVER"})
        await serverMock.startInstance()
        try:
            self.assertEqual(serverMock.lastClassId, "")
            self.assertEqual(serverMock.lastDeviceId, "")
            self.assertEqual(serverMock.lastConfigDouble, DEFAULT_LAST_DOUBLE)
            await instantiateFromName("ALICE", name="testConfig",
                                      serverId="MDL_SERVER")
            self.assertEqual(serverMock.lastClassId, "TestDevice")
            self.assertEqual(serverMock.lastDeviceId, "ALICE")
            self.assertEqual(serverMock.lastConfigDouble, 5.0)
        finally:
            await serverMock.slotKillDevice()

    @async_tst
    async def test_instantiate_device(self):
        serverMock = MockServer({"_deviceId_": "MDL_SERVER"})
        await serverMock.startInstance()
        try:
            h = Hash()
            h["deviceId"] = "ALICE"
            h["name"] = "testConfig"
            h["serverId"] = "MDL_SERVER"
            await call(MANAGER, "slotInstantiateDevice", h)
            self.assertEqual(serverMock.lastClassId, "TestDevice")
            self.assertEqual(serverMock.lastDeviceId, "ALICE")
            self.assertEqual(serverMock.lastConfigDouble, 5.0)

            # reset settings
            await serverMock.reset()
            self.assertEqual(serverMock.lastClassId, "")
            self.assertEqual(serverMock.lastDeviceId, "")
            self.assertEqual(serverMock.lastConfigDouble, DEFAULT_LAST_DOUBLE)

            h = Hash()
            h["deviceId"] = "ALICE"
            h["name"] = "testConfig"
            h["classId"] = "NoClass"
            h["serverId"] = "MDL_SERVER"
            # ClassId missmatch ...
            with self.assertRaises(KaraboError):
                await call(MANAGER, "slotInstantiateDevice", h)

            h = Hash()
            h["deviceId"] = "ALICE"
            h["serverId"] = "MDL_SERVER"
            await call(MANAGER, "slotInstantiateDevice", h)
            self.assertEqual(serverMock.lastClassId, "TestDevice")
            self.assertEqual(serverMock.lastDeviceId, "ALICE")
            self.assertEqual(serverMock.lastConfigDouble, 5.0)
        finally:
            await serverMock.slotKillDevice()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import os
import unittest
from datetime import datetime

from ..configuration_database import ConfigurationDatabase, DbHandle
from ..utils import (
    CONFIG_DB_CONFIG_SET_ID, CONFIG_DB_DATA, CONFIG_DB_DESCRIPTION,
    CONFIG_DB_DEVICE_ID, CONFIG_DB_DIFF_TIMEPOINT, CONFIG_DB_MAX_TIMEPOINT,
    CONFIG_DB_MIN_TIMEPOINT, CONFIG_DB_NAME, CONFIG_DB_OVERWRITABLE,
    CONFIG_DB_PRIORITY, CONFIG_DB_SCHEMA, CONFIG_DB_TIMEPOINT, CONFIG_DB_USER,
    ISO8601_FORMAT, ConfigurationDBError)

TEST_DB_PATH = 'configManagerTest.db'

CFG_NAME_1 = 'FirstConfig'
CFG_NAME_2 = 'SecondConfig'
CFG_NAME_3 = 'ThirdConfig'

CFG_DEVICE_1 = 'SCS/CAM/BASLER'
CFG_DEVICE_2 = 'CAS_LAB/DIGI/ADQ-12'
CFG_DEVICE_3 = 'KaraboProjectManager'

TIMESTAMP_1 = '2020-01-30T20:02:02.123000'
TIMESTAMP_2 = '2019-01-30T09:02:02.123000'


def create_config(deviceId):
    return {CONFIG_DB_DEVICE_ID: deviceId,
            CONFIG_DB_DATA: 'JFDHFYE947AGEUFHFSJDPOWIAKLSIDHD',
            CONFIG_DB_SCHEMA: 'QREPLCMBab789oeiryJAAHDDBAVDUUWY93948==='}


CFG_1 = [create_config(CFG_DEVICE_1)]
CFG_2 = [create_config(CFG_DEVICE_2)]
CFG_3 = [create_config(CFG_DEVICE_3)]

CFG_SET = [create_config(CFG_DEVICE_1), create_config(CFG_DEVICE_2)]
CFG_SET_2 = [
    create_config(CFG_DEVICE_1), create_config(CFG_DEVICE_2),
    create_config(CFG_DEVICE_3)]


class TestConfigurationManagerData(unittest.TestCase):

    @classmethod
    def setUp(cls):
        cls.connection = DbHandle(TEST_DB_PATH)
        cls.database = ConfigurationDatabase(cls.connection)
        cls.database.assureExisting()

    @classmethod
    def tearDown(cls):
        os.remove(cls.database.path)

    def test_save_get_config(self):
        """ Checks that saving one config and then retrieving it by its exact
            name works.
        """
        self.database.save_configuration(CFG_NAME_1, CFG_1)
        cfg = self.database.get_configuration(CFG_DEVICE_1, CFG_NAME_1)

        self.assertEqual(CFG_NAME_1, cfg['name'])
        self.assertEqual(CFG_1[0][CONFIG_DB_DATA], cfg[CONFIG_DB_DATA])
        self.assertEqual(CFG_1[0][CONFIG_DB_SCHEMA], cfg[CONFIG_DB_SCHEMA])
        self.assertEqual('', cfg[CONFIG_DB_DESCRIPTION])  # Empty default
        self.assertEqual('.', cfg[CONFIG_DB_USER])  # Default anonymous user
        self.assertEqual(1, cfg[CONFIG_DB_PRIORITY])  # Default priority, 1

        # Trying to retrieve with a non-existing name should return nothing.
        cfg = self.database.get_configuration(CFG_DEVICE_1,
                                              'Non-existing name')
        self.assertEqual(cfg, {})

    def test_save_get_config_name_part(self):
        """ Checks that saving a set of configs and then retrieving them by
            name parts works.
        """
        # Save with priority 1
        self.database.save_configuration(CFG_NAME_1, CFG_1,
                                         timestamp=TIMESTAMP_1)

        # CFG_SET contains one config for CFG_DEVICE_1 and one for CFG_DEVICE_2
        # in that order.
        self.database.save_configuration(CFG_NAME_2, CFG_SET,
                                         description='Bla bla',
                                         user='Bob',
                                         priority=3,
                                         timestamp=TIMESTAMP_2)

        cfg = self.database.list_configurations(CFG_DEVICE_1, CFG_NAME_2[:4])
        self.assertEqual(len(cfg), 1)
        self.assertEqual(CFG_NAME_2, cfg[0][CONFIG_DB_NAME])
        self.assertEqual(cfg[0][CONFIG_DB_TIMEPOINT], TIMESTAMP_2)
        self.assertEqual('Bla bla', cfg[0][CONFIG_DB_DESCRIPTION])
        self.assertEqual('Bob', cfg[0][CONFIG_DB_USER])
        self.assertEqual(3, cfg[0][CONFIG_DB_PRIORITY])
        # The config wasn't the only one saved by the save config call -
        # CFG_SET was used.
        self.assertNotEqual(-1, cfg[0][CONFIG_DB_CONFIG_SET_ID])

        devices_ids = [CFG_DEVICE_1, CFG_DEVICE_2]
        cfg = self.database.list_configuration_sets(devices_ids)
        self.assertEqual(len(cfg), 1)
        config_set = cfg[0]
        self.assertTrue(CONFIG_DB_DEVICE_ID not in config_set)
        self.assertEqual(config_set[CONFIG_DB_NAME], CFG_NAME_2)
        self.assertEqual(config_set[CONFIG_DB_MIN_TIMEPOINT], TIMESTAMP_2)
        self.assertEqual(config_set[CONFIG_DB_MAX_TIMEPOINT], TIMESTAMP_2)
        self.assertEqual(config_set[CONFIG_DB_DIFF_TIMEPOINT], 0.0)
        self.assertEqual(config_set[CONFIG_DB_DESCRIPTION], 'Bla bla')
        self.assertEqual(config_set[CONFIG_DB_USER], 'Bob')
        self.assertEqual(config_set[CONFIG_DB_PRIORITY], 3)

        cfg = self.database.list_configurations(CFG_DEVICE_1)
        # With an empty name part, all the configs for CFG_DEVICE_1 should be
        # retrieved
        self.assertEqual(len(cfg), 2)
        self.assertTrue(cfg[0][CONFIG_DB_NAME] in [CFG_NAME_1, CFG_NAME_2])
        self.assertTrue(cfg[1][CONFIG_DB_NAME] in [CFG_NAME_1, CFG_NAME_2])

        # A name part known to not happen in the data should retrieve nothing.
        cfg = self.database.list_configurations(CFG_DEVICE_1, 'inv part')
        self.assertEqual(cfg, [])

        # We should have stored 2 devices in the database
        devices = self.database.list_devices()
        self.assertIn(CFG_DEVICE_1, devices)
        self.assertIn(CFG_DEVICE_2, devices)
        self.assertEqual(len(devices), 2)

        devices = self.database.list_devices(priority=2)
        self.assertEqual(devices, [])

        devices = self.database.list_devices(priority=1)
        self.assertIn(CFG_DEVICE_1, devices)
        self.assertNotIn(CFG_DEVICE_2, devices)
        self.assertEqual(len(devices), 1)

    def test_save_get_last_config_priority(self):
        """ Checks that saving a set of configs and then retrieving the latest
            for a device with a given priority works.
        """
        self.database.save_configuration(CFG_NAME_1, CFG_SET,
                                         priority=2,
                                         timestamp=TIMESTAMP_2)

        # CFG_SET contains one config for CFG_DEVICE_1 and one for CFG_DEVICE_2
        # in that order. TIMESTAMP_1 is more recent than TIMESTAMP_2.
        self.database.save_configuration(CFG_NAME_2, CFG_SET,
                                         description='Bla bla',
                                         user='Bob',
                                         priority=1,
                                         timestamp=TIMESTAMP_2)

        self.database.save_configuration(CFG_NAME_3, CFG_SET,
                                         description='Bla bla',
                                         user='Bob',
                                         priority=1,
                                         timestamp=TIMESTAMP_1)

        cfg = self.database.get_last_configuration(CFG_DEVICE_1, priority=1)
        # For priority 1, the last config is the one saved under CFG_NAME_3.
        self.assertEqual(len(cfg), 8)
        self.assertEqual(CFG_NAME_3, cfg[CONFIG_DB_NAME])
        self.assertEqual(cfg[CONFIG_DB_TIMEPOINT], TIMESTAMP_1)
        self.assertEqual('Bla bla', cfg[CONFIG_DB_DESCRIPTION])
        self.assertEqual('Bob', cfg[CONFIG_DB_USER])
        self.assertEqual(1, cfg[CONFIG_DB_PRIORITY])
        self.assertEqual(False, cfg[CONFIG_DB_OVERWRITABLE])

        cfg = self.database.get_last_configuration(CFG_DEVICE_1, priority=2)
        # For priority 2, the last config is the one saved under CFG_NAME_1.
        self.assertEqual(len(cfg), 8)
        self.assertEqual(CFG_NAME_1, cfg[CONFIG_DB_NAME])
        self.assertEqual(cfg[CONFIG_DB_TIMEPOINT], TIMESTAMP_2)
        self.assertEqual('', cfg[CONFIG_DB_DESCRIPTION])
        self.assertEqual('.', cfg[CONFIG_DB_USER])
        self.assertEqual(2, cfg[CONFIG_DB_PRIORITY])
        self.assertEqual(False, cfg[CONFIG_DB_OVERWRITABLE])

        cfg = self.database.get_last_configuration(CFG_DEVICE_1, priority=3)
        # For priority 3, there's no config of CFG_DEVICE_1.
        self.assertEqual(cfg, {})

    def test_reject_config_repeat_name(self):
        """ Checks that an attempt to save a configuration with a name that is
            already in use by the same device fails.
        """
        self.database.save_configuration(CFG_NAME_1, CFG_1)
        with self.assertRaises(ConfigurationDBError) as verr:
            self.database.save_configuration(CFG_NAME_1, CFG_SET)
        self.assertTrue(verr.exception.args[0].find(
            'is already taken for at least') > -1)

    def test_config_name_taken(self):
        """ Checks that the is_config_name_taken returns true when there is at
            least one non-overwritable config for any of the devices in the
            list and if there's an overwritable config with the same name, it
            is for the exact same set of devices.
        """
        self.database.save_configuration(CFG_NAME_1, CFG_1)

        self.assertFalse(
            self.database.is_config_name_taken(
                CFG_NAME_1, [CFG_DEVICE_2]
            )
        )
        self.assertTrue(
            self.database.is_config_name_taken(CFG_NAME_1, [CFG_DEVICE_1])
        )
        self.assertTrue(
            self.database.is_config_name_taken(
                CFG_NAME_1,
                [CFG_DEVICE_1, CFG_DEVICE_2])
        )
        self.assertFalse(
            self.database.is_config_name_taken(
                CFG_NAME_2,
                [CFG_DEVICE_1, CFG_DEVICE_2])
        )

        # Checks for equally named overwritable configs for different sets of
        # devices - that condition must also trigger a name already taken
        # error.
        self.database.save_configuration(
            CFG_NAME_3, CFG_SET, overwritable=True)
        with self.assertRaises(ConfigurationDBError) as verr:
            self.database.save_configuration(
                CFG_NAME_3, CFG_1, overwritable=False
            )
        self.assertTrue(verr.exception.args[0].find(
            'is already taken for at least') > -1)
        with self.assertRaises(ConfigurationDBError) as verr:
            self.database.save_configuration(
                CFG_NAME_3, CFG_1, overwritable=True
            )
        self.assertTrue(verr.exception.args[0].find(
            'is already taken for at least') > -1)

    def test_device_conf_limit(self):
        from .. import configuration_database as dbsettings
        dbsettings.CONFIGURATION_LIMIT = 20
        deviceName = "DeviceExceeds"

        config = [create_config(deviceName)]
        for i in range(dbsettings.CONFIGURATION_LIMIT):
            config_name = f"exceed-{i}"
            self.database.save_configuration(config_name, config)

        with self.assertRaises(ConfigurationDBError):
            self.database.save_configuration("exceed-crash", config)

    def test_overwritable_conf(self):
        """ Checks that an overwritable configuration can be overwritten,
            even at the edge of configuration limit per device. Also checks
            that ovewriting a configuration without specifying a timestamp
            updates the timestamp to the current time.
        """
        from .. import configuration_database as dbsettings
        dbsettings.CONFIGURATION_LIMIT = 10
        deviceName = "DeviceExceeds-WithOverwritableConfig"

        config = [create_config(deviceName)]
        for i in range(dbsettings.CONFIGURATION_LIMIT - 1):
            config_name = f"exceed-{i}"
            self.database.save_configuration(config_name, config)

        # Saves an overwritable configuration at the limit.
        overCfg = [create_config(deviceName)]
        overCfgName = f"exceed-over-{dbsettings.CONFIGURATION_LIMIT}"
        self.database.save_configuration(
            overCfgName, overCfg, overwritable=True)

        cfg = self.database.get_configuration(deviceName, overCfgName)
        self.assertEqual(len(cfg), 8)
        self.assertEqual(overCfgName, cfg[CONFIG_DB_NAME])
        self.assertEqual('', cfg[CONFIG_DB_DESCRIPTION])
        self.assertEqual('.', cfg[CONFIG_DB_USER])
        self.assertEqual(1, cfg[CONFIG_DB_PRIORITY])
        self.assertEqual(True, cfg[CONFIG_DB_OVERWRITABLE])
        timeStamp1 = datetime.strptime(cfg[CONFIG_DB_TIMEPOINT],
                                       ISO8601_FORMAT)

        # Ovewrites the configuration at the limit.
        # Overwritable flag must not be updated.
        # Priority, description and timestamp must be updated.
        self.database.save_configuration(
            overCfgName, overCfg, description="BlahBlah",
            priority=3, overwritable=False
        )
        cfg = self.database.get_configuration(deviceName, overCfgName)
        self.assertEqual(len(cfg), 8)
        self.assertEqual(overCfgName, cfg[CONFIG_DB_NAME])
        self.assertEqual('BlahBlah', cfg[CONFIG_DB_DESCRIPTION])
        self.assertEqual('.', cfg[CONFIG_DB_USER])
        self.assertEqual(3, cfg[CONFIG_DB_PRIORITY])
        self.assertEqual(True, cfg[CONFIG_DB_OVERWRITABLE])
        timeStamp2 = datetime.strptime(cfg[CONFIG_DB_TIMEPOINT],
                                       ISO8601_FORMAT)
        self.assertGreater(timeStamp2, timeStamp1)

        # Overwrite the configuration a second time, to be
        # sure that all the device configurations will be properly
        # updated.
        self.database.save_configuration(
            overCfgName, overCfg, description="Bleh-Bleh", priority=2)
        cfg = self.database.get_configuration(deviceName, overCfgName)
        self.assertEqual(len(cfg), 8)
        self.assertEqual(overCfgName, cfg[CONFIG_DB_NAME])
        self.assertEqual('Bleh-Bleh', cfg[CONFIG_DB_DESCRIPTION])
        self.assertEqual('.', cfg[CONFIG_DB_USER])
        self.assertEqual(2, cfg[CONFIG_DB_PRIORITY])
        self.assertEqual(True, cfg[CONFIG_DB_OVERWRITABLE])
        timeStamp3 = datetime.strptime(cfg[CONFIG_DB_TIMEPOINT],
                                       ISO8601_FORMAT)
        self.assertGreater(timeStamp3, timeStamp2)

        # An attempt to save a new configuration, overwritable or not,
        # should fail once the limit is reached.
        with self.assertRaises(ConfigurationDBError):
            self.database.save_configuration(
                "exceed-crash",  overCfg, f"{overCfgName}-limit",
                overwritable=True)

    def test_configSets_minSize(self):
        """Checks that retrieving configuration sets specifying a minSetSize
        that is lower than the size of the original configuration sets allows
        retrieving config sets with a subset of the devices specified. Also
        checks that retrieving configuration sets omitting the minSetSize
        retrieves a superset of the devices specified.
        """
        # Save with priority 1
        self.database.save_configuration(CFG_NAME_1, CFG_1,
                                         user='Alice',
                                         timestamp=TIMESTAMP_1)

        # CFG_SET contains one config for CFG_DEVICE_1 and one for CFG_DEVICE_2
        # in that order.
        self.database.save_configuration(CFG_NAME_2, CFG_SET,
                                         description='Bla bla',
                                         user='Bob',
                                         priority=3,
                                         timestamp=TIMESTAMP_2)

        # CFG_SET_2 contains three configs: one for CFG_DEVICE_1, one for
        # CFG_DEVICE_2 and one for CFG_DEVICE_3.
        self.database.save_configuration(CFG_NAME_3, CFG_SET_2,
                                         user='Charlie')

        # Retrieval of all sets of size at least 1 containing CFG_DEVICE_1
        # should return all three saved sets.
        cfgs = self.database.list_configuration_sets([CFG_DEVICE_1], 1)

        self.assertEqual(len(cfgs), 3)
        self.assertEqual(cfgs[0][CONFIG_DB_NAME], CFG_NAME_1)
        self.assertEqual(cfgs[0][CONFIG_DB_USER], 'Alice')
        # list_configurations_sets returns the config set id even for single
        # configuration sets.
        self.assertNotEqual(cfgs[0][CONFIG_DB_CONFIG_SET_ID], -1)

        self.assertEqual(cfgs[1][CONFIG_DB_NAME], CFG_NAME_2)
        self.assertEqual(cfgs[1][CONFIG_DB_USER], 'Bob')
        self.assertNotEqual(cfgs[1][CONFIG_DB_CONFIG_SET_ID], -1)

        self.assertEqual(cfgs[2][CONFIG_DB_NAME], CFG_NAME_3)
        self.assertEqual(cfgs[2][CONFIG_DB_USER], 'Charlie')
        self.assertNotEqual(cfgs[2][CONFIG_DB_CONFIG_SET_ID], -1)

        # Checks the configs in the config set for CFG_NAME_2
        cfgs_in_set = self.database.list_configurations_in_set(
            cfgs[1][CONFIG_DB_CONFIG_SET_ID])
        self.assertEqual(len(cfgs_in_set), 2)
        # Config set with CFG_NAME_2 has CFG_DEVICE_1 and CFG_DEVICE_2
        devs = {CFG_DEVICE_1, CFG_DEVICE_2}
        self.assertTrue(cfgs_in_set[0][CONFIG_DB_DEVICE_ID] in devs)
        self.assertTrue(cfgs_in_set[1][CONFIG_DB_DEVICE_ID] in devs)
        self.assertEqual(cfgs_in_set[0][CONFIG_DB_TIMEPOINT], TIMESTAMP_2)
        self.assertEqual(cfgs_in_set[1][CONFIG_DB_TIMEPOINT], TIMESTAMP_2)

        # Retrieval of all sets of unspecified size containing CFG_DEVICE_1
        # should return all three saved sets - when the value is not specified,
        # the number of devicesIds in the argument list is used; so this is
        # the same call of the previous case.
        cfgs = self.database.list_configuration_sets([CFG_DEVICE_1])
        self.assertEqual(len(cfgs), 3)
        self.assertEqual(cfgs[0][CONFIG_DB_NAME], CFG_NAME_1)
        self.assertEqual(cfgs[0][CONFIG_DB_USER], 'Alice')
        self.assertEqual(cfgs[1][CONFIG_DB_NAME], CFG_NAME_2)
        self.assertEqual(cfgs[1][CONFIG_DB_USER], 'Bob')
        self.assertEqual(cfgs[2][CONFIG_DB_NAME], CFG_NAME_3)
        self.assertEqual(cfgs[2][CONFIG_DB_USER], 'Charlie')
        # minSetSize=0 or negative mean the same as the default None
        cfgs2 = self.database.list_configuration_sets([CFG_DEVICE_1], 0)
        self.assertEqual(cfgs, cfgs2)
        cfgs2 = self.database.list_configuration_sets([CFG_DEVICE_1], -4)
        self.assertEqual(cfgs, cfgs2)

        # Retrieval of all sets of size at least 2 containing both CFG_DEVICE_1
        # and CFG_DEVICE_2 should return both config sets that contained both
        # devices.
        cfgs = self.database.list_configuration_sets(
            [CFG_DEVICE_1, CFG_DEVICE_2], 2)
        self.assertEqual(len(cfgs), 2)
        self.assertEqual(cfgs[0][CONFIG_DB_NAME], CFG_NAME_2)
        self.assertEqual(cfgs[0][CONFIG_DB_USER], 'Bob')
        self.assertEqual(cfgs[1][CONFIG_DB_NAME], CFG_NAME_3)
        self.assertEqual(cfgs[1][CONFIG_DB_USER], 'Charlie')
        # Retrieval of all sets of size at least 1 containing at least
        # CFG_DEVICE_2 or CFG_DEVICE_3 should return the two multi-device set,
        # one for having CFG_DEVICE_2 and the other for having both devices.
        cfgs = self.database.list_configuration_sets(
            [CFG_DEVICE_2, CFG_DEVICE_3], 1)
        self.assertEqual(len(cfgs), 2)
        self.assertEqual(cfgs[0][CONFIG_DB_NAME], CFG_NAME_2)
        self.assertEqual(cfgs[0][CONFIG_DB_USER], 'Bob')
        self.assertEqual(cfgs[1][CONFIG_DB_NAME], CFG_NAME_3)
        self.assertEqual(cfgs[1][CONFIG_DB_USER], 'Charlie')
        # Retrieval of all sets of size at least 2 containing CFG_DEVICE_1,
        # CFG_DEVICE_2 or CFG_DEVICE_3 should return both config sets that have
        # more than one device.
        cfgs = self.database.list_configuration_sets(
            [CFG_DEVICE_1, CFG_DEVICE_2, CFG_DEVICE_3], 2)
        self.assertEqual(len(cfgs), 2)
        self.assertEqual(cfgs[0][CONFIG_DB_NAME], CFG_NAME_2)
        self.assertEqual(cfgs[0][CONFIG_DB_USER], 'Bob')
        self.assertEqual(cfgs[1][CONFIG_DB_NAME], CFG_NAME_3)
        self.assertEqual(cfgs[1][CONFIG_DB_USER], 'Charlie')
        # Retrieval of all sets of size at least 3 containing CFG_DEVICE_1,
        # CFG_DEVICE_2 and CFG_DEVICE_3 should return just the config set that
        # has the three devices (not specifying minSetSize in the same as
        # specifying len(deviceIds)).
        cfgs = self.database.list_configuration_sets(
            [CFG_DEVICE_1, CFG_DEVICE_2, CFG_DEVICE_3])
        self.assertEqual(len(cfgs), 1)
        self.assertEqual(cfgs[0][CONFIG_DB_NAME], CFG_NAME_3)
        self.assertEqual(cfgs[0][CONFIG_DB_USER], 'Charlie')
        # Retrieval of all sets of size at least 1 containing CFG_DEVICE_3
        # should return only the device set that had the device and two
        # more - this is the only device set that has CFG_DEVICE_3.
        cfgs = self.database.list_configuration_sets([CFG_DEVICE_3], 1)
        self.assertEqual(len(cfgs), 1)
        self.assertEqual(cfgs[0][CONFIG_DB_NAME], CFG_NAME_3)
        self.assertEqual(cfgs[0][CONFIG_DB_USER], 'Charlie')
        # Retrieval of all sets of size at least 4 containing CFG_DEVICE_3
        # should throw an exception - minSetSize cannot be greater than the
        # number of devices in the first argument.
        with self.assertRaises(ConfigurationDBError):
            cfgs = self.database.list_configuration_sets([CFG_DEVICE_3], 4)
        # Empty list of devices throws exceptions.
        with self.assertRaises(ConfigurationDBError):
            cfgs = self.database.list_configuration_sets([])

    def test_save_get_configurations_in_set(self):
        """Checks that list configurations for device (independent of name
        part specification) returns information about the config set as part
        of which that configuration has been saved, and that all the configs
        in the set can be easily retrieved by list configuration for
        config set.
        """
        # Creates a config set with a single configuration for CFG_DEVICE_1
        self.database.save_configuration(CFG_NAME_1, CFG_1,
                                         user='Alice',
                                         timestamp=TIMESTAMP_1)

        # CFG_SET contains one config for CFG_DEVICE_1 and one for CFG_DEVICE_2
        # in that order.
        self.database.save_configuration(CFG_NAME_2, CFG_SET,
                                         description='Bla bla',
                                         user='Bob',
                                         priority=3,
                                         timestamp=TIMESTAMP_1)

        # CFG_SET_2 contains three configs: one for CFG_DEVICE_1, one for
        # CFG_DEVICE_2 and one for CFG_DEVICE_3.
        self.database.save_configuration(CFG_NAME_3, CFG_SET_2,
                                         user='Charlie',
                                         timestamp=TIMESTAMP_2)

        # Gets all the configs for CFG_DEVICE_1 - all the three generated
        # config sets have one config for CFG_DEVICE_1.
        dev1_cfgs = self.database.list_configurations(CFG_DEVICE_1)
        self.assertEqual(len(dev1_cfgs), 3)
        # The first CFG_DEVICE_1 config has been saved individually.
        self.assertEqual(dev1_cfgs[0][CONFIG_DB_CONFIG_SET_ID], -1)
        # The other two CFG_DEVICE_1 configs have been saved as part of
        # multi-device config sets
        self.assertNotEqual(dev1_cfgs[1][CONFIG_DB_CONFIG_SET_ID], -1)
        self.assertNotEqual(dev1_cfgs[2][CONFIG_DB_CONFIG_SET_ID], -1)
        # Checks that the configurations in each of the multi-device config
        # sets are correct.
        cfgs = self.database.list_configurations_in_set(
            dev1_cfgs[1][CONFIG_DB_CONFIG_SET_ID])
        self.assertEqual(len(cfgs), 2)
        devs = {CFG_DEVICE_1, CFG_DEVICE_2}
        self.assertTrue(cfgs[0][CONFIG_DB_DEVICE_ID] in devs)
        self.assertTrue(cfgs[1][CONFIG_DB_DEVICE_ID] in devs)
        self.assertEqual(cfgs[0][CONFIG_DB_TIMEPOINT], TIMESTAMP_1)
        self.assertEqual(cfgs[1][CONFIG_DB_TIMEPOINT], TIMESTAMP_1)
        cfgs = self.database.list_configurations_in_set(
            dev1_cfgs[2][CONFIG_DB_CONFIG_SET_ID])
        devs.add(CFG_DEVICE_3)
        self.assertEqual(len(cfgs), 3)
        self.assertTrue(cfgs[0][CONFIG_DB_DEVICE_ID] in devs)
        self.assertTrue(cfgs[1][CONFIG_DB_DEVICE_ID] in devs)
        self.assertTrue(cfgs[2][CONFIG_DB_DEVICE_ID] in devs)
        self.assertEqual(cfgs[0][CONFIG_DB_TIMEPOINT], TIMESTAMP_2)
        self.assertEqual(cfgs[1][CONFIG_DB_TIMEPOINT], TIMESTAMP_2)
        self.assertEqual(cfgs[2][CONFIG_DB_TIMEPOINT], TIMESTAMP_2)


if __name__ == '__main__':
    unittest.main()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import ast
import collections
import functools
import logging
import os
import uuid
import weakref
from asyncio import get_running_loop, sleep, wait_for
from contextlib import contextmanager
from pathlib import Path
from typing import Any

import pytest

from ..device_client import getProperties
from ..eventloop import synchronize_notimeout


def get_ast_objects(package, ignore=[]):
    """Get all ast objects from a specified package

    :param ignore: list of filenames that are ignored in the lookup
    """
    def _get_ast(path):
        """Get an ast.AST object for the specified file"""
        with open(path, 'rb') as fp:
            return compile(fp.read(), path, "exec", ast.PyCF_ONLY_AST)

    common_dir = str(Path(package.__file__).parent)
    ast_objects = []
    for dirpath, _, filenames in os.walk(common_dir):
        for fn in filenames:
            if Path(fn).suffix == ".py" and fn not in ignore:
                path = str(Path(dirpath).joinpath(fn))
                ast_objects.append(_get_ast(path))
    return ast_objects


def run_test(func):
    """Run a pytest test function `func` in the karabo eventloop"""

    @functools.wraps(func)
    @pytest.mark.asyncio(loop_scope="module")
    async def wrapper(*args, **kwargs):
        loop = get_running_loop()
        lead = getattr(loop, "lead", None)
        task = loop.create_task(loop.run_coroutine_or_thread(
            func, *args, **kwargs), instance=lead)
        return await task

    return wrapper


_LoggingWatcher = collections.namedtuple("_LoggingWatcher",
                                         ["records", "output"])


class _CapturingHandler(logging.Handler):
    """A logging handler capturing all (raw and formatted) logging output.
    """

    def __init__(self):
        logging.Handler.__init__(self)
        self.watcher = _LoggingWatcher([], [])

    def flush(self):
        pass

    def emit(self, record):
        self.watcher.records.append(record)
        msg = self.format(record)
        self.watcher.output.append(msg)


@contextmanager
def assertLogs(logger_name=None, level=None):
    """A context manager used to implement assertLogs().

    :param logger_name: defaults to `None` (root)
    :param level: The logging level, defaults to `None` -> INFO
    """
    try:
        if isinstance(logger_name, logging.Logger):
            logger = logger_name
        else:
            logger = logging.getLogger(logger_name)

        if level:
            level = logging._nameToLevel.get(level, level)
        else:
            level = logging.INFO

        # Store old
        old_handlers = logger.handlers[:]
        old_level = logger.level
        old_propagate = logger.propagate
        # Attach new
        LOGGING_FORMAT = "%(levelname)s:%(name)s:%(message)s"
        formatter = logging.Formatter(LOGGING_FORMAT)
        handler = _CapturingHandler()
        handler.setFormatter(formatter)
        watcher = handler.watcher
        logger.handlers = [handler]
        logger.setLevel(level)
        logger.propagate = False

        yield handler.watcher
    finally:
        logger.handlers = old_handlers
        logger.propagate = old_propagate
        logger.setLevel(old_level)

        if len(watcher.records) == 0:
            raise AssertionError(
                "no logs of level {} or higher triggered on {}"
                .format(logging.getLevelName(level), logger.name))


@synchronize_notimeout
async def sleepUntil(condition, timeout=None):
    """Sleep until some condition is valid

    :param timeout: The timeout parameter, defaults to `None`. (no timeout)
    """

    async def internal_wait():
        while not condition():
            await sleep(0.05)

    return await wait_for(internal_wait(), timeout=timeout)


def create_instanceId(name=None):
    """Create a unique instanceId with `name` and append a uuid"""
    if name is None:
        name = "test-mdl"
    return f"{name}-{uuid.uuid4()}"


@contextmanager
def switch_instance(instance):
    """Switch the owning instance of the loop"""
    loop = get_running_loop()
    try:
        loop_instance = loop.instance
        loop.instance = weakref.ref(instance)
        yield
    finally:
        loop.instance = loop_instance


async def assert_wait_property(
        deviceId: str, path: str, value: Any, timeout: int = 5):
    """Asserts the successful wait for a property with `path`.

    We timeout after timeout seconds.
    """
    POLL_WAIT = 0.5
    current = None

    async def internal_wait():
        nonlocal current

        while True:
            try:
                h = await wait_for(getProperties(deviceId, path), POLL_WAIT)
            except TimeoutError:
                continue
            current = h[path]
            if current == value:
                break
            else:
                await sleep(POLL_WAIT)

    try:
        await wait_for(internal_wait(), timeout=timeout)
    except TimeoutError:
        text = f"Waiting for path {path}, {value} on {deviceId}: {current}"
        assert False, text



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa: E501
# this file contains xquery code
import os
from textwrap import dedent
from time import gmtime, strftime, strptime

from pyexistdb.exceptions import ExistDBException

from ..bases import DatabaseBase
from .dbsettings import DbSettings
from .util import LIST_DOMAINS_QUERY, ProjectDBError, assure_running

DATE_FORMAT = '%Y-%m-%d %H:%M:%S'


class ProjectDatabase(DatabaseBase):

    def __init__(self, user, password, server=None, port=None,
                 test_mode=False, init_db=False):
        """
        Create a project data base context for a given user
        :param user: the user, can be either admin for local db, or set on the
                     remote projectDB
        :param password: user's password. Is 'karabo' in local context
        :param server: server to connect to. If left blank the
           'KARABO_PROJECT_DB' environment variable will be used
        :param port: the port which to connect to, defaults to 8080 or the
           'KARABO_PROJECT_DB_PORT' environment variable if present
        :param test_mode: defaults to False. In this case
           db_settings.root_collection will be used as the root collection,
           otherwise, root_collection_test is used.
           If test_mode is True, the server and port options will be evaluated
           from the KARABO_TEST_PROJECT_DB and KARABO_TEST_PROJECT_DB_PORT
           variables respectively.
        :param init_db: defaults to False. If True, the default collections
           will be added if missing on context entry.
        :return: a ProjectDatabase context
        """
        super().__init__()
        if test_mode:
            server = os.getenv('KARABO_TEST_PROJECT_DB', 'localhost')
            port = os.getenv('KARABO_TEST_PROJECT_DB_PORT', 8181)

        # get our environment straightened out
        if server is None:
            server = os.getenv('KARABO_PROJECT_DB', None)

        if server is None:
            raise OSError("No environment variable KARABO_PROJECT_DB"
                          " found, nor was a server to connect to"
                          " given!")
        if port is None:
            port = os.getenv('KARABO_PROJECT_DB_PORT', 8181)

        port = int(port)
        self.settings = DbSettings(user, password, server, port, init_db)
        self.root = (self.settings.root_collection if not test_mode else
                     self.settings.root_collection_test)

    def onEnter(self):
        return assure_running(self.settings)

    def _check_for_modification(self, domain, uuid, old_date):
        """ Check whether the item with of the given `domain` and `uuid` was
        modified

        :param domain: the domain under which this item exists
        :param uuid: the item's uuid
        :param old_date: the item's last modified time stamp
        :return A tuple stating whether the item was modified inbetween and a
                string describing the reason
        """
        # Check whether the object got modified inbetween
        path = f"{self.root}/{domain}"
        query = """
            xquery version "3.0";
            let $path := "{path}"
            let $uuid := "{uuid}"

            for $doc in collection($path)/xml[@uuid = $uuid]
            let $date := $doc/@date
            return <item date="{{$date}}"/>
                """.format(path=path, uuid=uuid)
        modified = False
        reason = ""
        try:
            res = self.dbhandle.query(query)
            if res.results:
                current_date = res.results[0].get('date')
            else:
                # Item not yet existing
                return modified, reason
        except ExistDBException as e:
            raise ProjectDBError(e)

        current_modified = strptime(current_date, DATE_FORMAT)
        last_modified = strptime(old_date, DATE_FORMAT)
        if last_modified < current_modified:
            # Item got saved inbetween
            modified = True
            reason = "Versioning conflict! Document modified inbetween."
        return modified, reason

    def load_item(self, domain, items):
        """
        Load an item or items from `domain`
        :param domain: a domain to load from
        :param item: the name of the item(s) to load
        :return:
        """
        # path to where the possible entries are located
        path = f"{self.root}/{domain}"

        assert isinstance(items, (list, tuple))

        results = []
        n_items = len(items)
        # we re-chunk the request for querying as everything ends up in a
        # single get call to the restful API. Through trial 50 seems a
        # reasonable trade-off between size of query and return value size.
        req_cnk_size = 50

        for i in range(0, n_items, req_cnk_size):
            max_idx = min(i + req_cnk_size, n_items)
            uuids = '","'.join(items[i:max_idx])
            query = """
            xquery version "3.0";

            let $uuids := {uuids}
            let $path := "{path}"

            for $uuid in $uuids
            let $sorted-docs :=
                for $doc in collection($path)/xml[@uuid = $uuid]
                order by $doc/@date
                return $doc
            return $sorted-docs[last()]

            """.format(uuids=f'("{uuids}")', path=path)

            try:
                res = self.dbhandle.query(query, how_many=req_cnk_size)
                for r in res.results:
                    results.append({'uuid': r.get('uuid'),
                                    'xml': self._make_str_if_needed(r)})
            except ExistDBException as e:
                raise ProjectDBError(e)

        return results

    def list_items(self, domain, item_types=None):
        """
        List items in domain which match item_types if given, or all items
        if not given
        :param domain: domain to list items from
        :param item_types: list or tuple of item_types to list
        :return: a list of dicts where each entry has keys: uuid, item_type
                 and simple_name
        """
        # path to where the possible entries are located
        path = f"{self.root}/{domain}"
        query = """
        xquery version "3.0";
        declare namespace functx = "http://www.functx.com";
        declare function functx:if-absent(
            $arg as item()* , $value as item()*)  as item()*
        {{ if (exists($arg)) then $arg else $value }};

        let $path := "{path}"
        {maybe_let}
        return <items>{{
        for $doc in collection($path)/xml{maybe_where}
        let $uuid := $doc/@uuid
        let $simple_name := $doc/@simple_name
        let $item_type := $doc/@item_type
        let $is_trashed := functx:if-absent($doc/@is_trashed, 'false')
        let $user := functx:if-absent($doc/@user, '')
        let $date := functx:if-absent($doc/@date, '')
        let $description := functx:if-absent($doc/@description, '')
        group by $uuid, $simple_name, $item_type, $is_trashed, $user, $date,
        $description
        return <item uuid="{{$uuid}}"
                simple_name="{{$simple_name}}"
                item_type="{{$item_type}}"
                is_trashed="{{$is_trashed}}"
                user="{{$user}}"
                date="{{$date}}"
                description="{{$description}}" />
        }}</items>
        """
        maybe_let, maybe_where = '', ''
        if item_types is not None:
            maybe_let_tmp = "let $item_types := ('{}')"
            maybe_let = maybe_let_tmp.format("','".join(item_types))
            maybe_where = '[@item_type=$item_types]'

        query = query.format(maybe_let=maybe_let,
                             maybe_where=maybe_where,
                             path=path)
        try:
            res = self.dbhandle.query(query)
            return [{'uuid': r.attrib['uuid'],
                     'item_type': r.attrib['item_type'],
                     'simple_name': r.attrib['simple_name'],
                     'is_trashed': r.attrib['is_trashed'],
                     'user': r.attrib['user'],
                     'date': r.attrib['date'],
                     'description': r.attrib['description']}
                    for r in res.results[0].getchildren()]
        except ExistDBException as e:
            raise ProjectDBError(e)

    def list_named_items(self, domain, item_type, simple_name):
        """
        List items in domain which match item_type and simple_name

        :param domain: domain to list items from
        :param item_type: item_type to match
        :param simple_name: iterable of names to match
        :return: a list of dicts where each entry has keys: uuid, item_type
                 and simple_name
        """
        # path to where the possible entries are located
        path = f"{self.root}/{domain}"
        query = """
        xquery version "3.0";
        declare namespace functx = "http://www.functx.com";
        declare function functx:if-absent(
            $arg as item()* , $value as item()*)  as item()*
        {{ if (exists($arg)) then $arg else $value }};

        let $path := "{path}"
        return <items>{{
        for $doc in collection($path)/xml{where}
        let $uuid := $doc/@uuid
        let $simple_name := $doc/@simple_name
        let $item_type := $doc/@item_type
        let $is_trashed := functx:if-absent($doc/@is_trashed, 'false')
        let $user := functx:if-absent($doc/@user, '')
        let $date := functx:if-absent($doc/@date, '')
        let $description := functx:if-absent($doc/@description, '')
        group by $uuid, $simple_name, $item_type, $is_trashed, $user, $date,
        $description
        return <item uuid="{{$uuid}}"
                simple_name="{{$simple_name}}"
                item_type="{{$item_type}}"
                is_trashed="{{$is_trashed}}"
                user="{{$user}}"
                date="{{$date}}"
                description="{{$description}}" />
        }}</items>
        """
        where = "[@item_type=('{}') and @simple_name=('{}')]".format(
            item_type, simple_name)

        query = query.format(where=where,
                             path=path)
        try:
            res = self.dbhandle.query(query)
            return [{'uuid': r.attrib['uuid'],
                     'item_type': r.attrib['item_type'],
                     'simple_name': r.attrib['simple_name'],
                     'is_trashed': r.attrib['is_trashed'],
                     'user': r.attrib['user'],
                     'date': r.attrib['date'],
                     'description': r.attrib['description']}
                    for r in res.results[0].getchildren()]
        except ExistDBException as e:
            raise ProjectDBError(e)

    def list_domains(self):
        """
        List top level domains in database
        :return:
        """
        query = LIST_DOMAINS_QUERY.format(self.root)
        try:
            res = self.dbhandle.query(query)
            return [r.text for r in res.results[0]]
        except ExistDBException:
            return []

    def sanitize_database(self, domain):
        """ This function (attempts to) sanitize all items in a domain

        NOTE: to run it
        from karabo.project_db.project_database import ProjectDatabase
        db = ProjectDatabase("karabo", "karabo", server="exflst105")
        with db as db:
            r = db.sanitize_database("/db/krb_config/CAS_INTERNAL")

        It inserts the minimally expected attributes if missing! A backup of
        the collection is created prior to the operation!
        """

        domain = domain.rstrip('/')
        tstamp = strftime("%Y-%m-%d_%H%M%S", gmtime())
        origin_base = '/'.join(domain.split('/')[:-1]) + "/"
        base = '{}{}/'.format('/'.join(domain.split('/')[:-2]),
                              self.settings.root_collection_backup)
        source = domain.split('/')[-1]
        bck = f"{source}_{tstamp}_backup"
        tmp = f"tmp_{tstamp}"

        query = """
            xquery version "3.0";
            import module namespace xmldb="http://exist-db.org/xquery/xmldb";

            let $origin_base := "{origin_base}"
            let $base := "{base}"
            let $col1 := "{source}"
            let $tmp := "{tmp}"
            let $col2 := "{dest}"

            let $res := xmldb:create-collection($base, $tmp)
            let $cpy_res := xmldb:copy(concat($origin_base,$col1), concat($base,$tmp))
            let $rnm_res := xmldb:rename(concat($base,$tmp,"/",$col1), $col2)
            let $mv_res := xmldb:move(concat($base,$tmp,"/",$col2), $base)
            return xmldb:remove(concat($base,$tmp))

        """.format(origin_base=origin_base, base=base, source=source, tmp=tmp,
                   dest=bck)

        print(f"Creating backup of domain {domain} at {base}{bck}")
        try:
            res = self.dbhandle.query(query)
        except ExistDBException as e:
            raise ProjectDBError(f"Failed creating backup: {e}")

        print("Sucessfully created backup, now proceeding with sanitizing")

        query = """
            xquery version "3.0";

            let $col := "{domain}"

            (: This query updates root elements missing aliases :)
            let $root_updates := for $doc in collection($col)/xml[not(@alias)]
            let $update := update insert attribute alias {{'default'}} into $doc
            return $doc

            (: This query updates project elements missing revisions :)
            let $docs := for $doc in collection($col)/xml[@item_type='project']
            where exists($doc/*/*/KRB_Item) and (not(exists($doc/*/*/KRB_Item/revision)))
            return $doc

            let $project_updates := for $doc in $docs
            let $update := update insert <revision KRB_Type="INT32">0</revision> into $doc/*/*/KRB_Item
            return $doc


            (: This query updates device server elements missing revisions :)
            let $device_server_updates := for $doc in collection($col)/xml[@item_type='device_server']/device_server/device_instance[not(@revision)]
            let $update := update insert attribute revision {{'0'}} into $doc
            return $doc


            (: These queries updates device config elements missing revisions :)
            let $config_updates_1 := for $doc in collection($col)/xml[@item_type='device_instance']/device_instance/device_config[not(@revision)]
            let $update := update insert attribute revision {{'0'}} into $doc
            return $doc

            let $config_updates_2 := for $doc in collection($col)/xml[@item_type='device_instance'][not(@active_rev)]
            let $update := update insert attribute active_rev {{'0'}} into $doc
            return $doc

            return <updates>
                <root_updates>{{count($root_updates)}}</root_updates>
                <project_updates>{{count($project_updates)}}</project_updates>
                <device_server_updates>{{count($device_server_updates)}}</device_server_updates>
                <config_updates_1>{{count($config_updates_1)}}</config_updates_1>
                <config_updates_2>{{count($config_updates_2)}}</config_updates_2>
                </updates>

            """.format(domain=domain)

        try:
            res = self.dbhandle.query(query).results[0]
        except ExistDBException as e:
            raise ProjectDBError(f"Failed sanitizing: {e}")

        msg = dedent("""
            Sanitized database:
            -------------------
            Root elements updated:                   {root_updates}
            Project elements updated:                {project_updates}
            Device server elements updated:          {device_server_updates}
            Configuration elements updated (root):   {config_updates_1}
            Configuration elements updated (child):  {config_updates_2}
        """.format(root_updates=res[0].text,
                   project_updates=res[1].text,
                   device_server_updates=res[2].text,
                   config_updates_1=res[3].text,
                   config_updates_2=res[4].text))
        print(msg)

    def update_attributes(self, items):
        """ Update attribute for the given ``items``

        :param items: list of Hashes containing information on which items
                      to update. Each list entry should be a Hash containing

                      - domain: domain the item resides at
                      - uuid: the uuid of the item
                      - item_type: indicate type of item which attribute should
                                   be changed
                      - attr_name: name of attribute which should be changed
                      - attr_value: value of attribute which should be changed

        :return: a list of dicts where each entry has keys: domain, uuid,
                 item_type, attr_name, attr_value
        """
        res_items = []
        for it in items:
            domain = it.get('domain')
            item_type = it.get('item_type')
            uuid = it.get('uuid')
            attr_name = it.get('attr_name')
            attr_value = it.get('attr_value')
            path = self.path(domain, uuid)
            query = """
                xquery version "3.0";

                let $doc := doc("{path}")/xml[@item_type="{item_type}"]
                let $cond :=
                    if (exists($doc/@{attr_name}))
                    then update value $doc/@{attr_name} with "{attr_value}"
                    else update insert attribute {attr_name} {{"{attr_value}"}}
                         into $doc

                return <doc uuid="{{$doc/@uuid}}"
                        item_type="{{$doc/@item_type}}"
                        attr_name="{attr_name}"
                        attr_value="{{$doc/@{attr_name}}}"/>
                    """.format(path=path, item_type=item_type,
                               attr_name=attr_name, attr_value=attr_value)

            try:
                res = self.dbhandle.query(query)
                root = res.results[0]
                res_items.append({'domain': domain,
                                  'uuid': root.attrib['uuid'],
                                  'item_type': root.attrib['item_type'],
                                  'attr_name': root.attrib['attr_name'],
                                  'attr_value': root.attrib['attr_value']})
            except ExistDBException as e:
                raise ProjectDBError(f"Failed updating attribute: {e}")
        return res_items

    def get_configurations_from_device_name(self, domain, instance_id):
        """
        Returns a list of configurations for a given device
        :param domain: DB domain
        :param instance_id: instance id of the device
        :return: a list of dicts:
            [{"configid": uuid of the configuration,
              "instanceid": device instance uuid in the DB},
              ...
            ]
        """
        query = """
        xquery version "3.0";
        let $path := "{path}"
        let $iid := "{instance_id}"
        return <items>{{
        for $doc in collection($path)/xml//device_instance[@instance_id=$iid]
        let $active := $doc/@active_rev
        let $configid := $doc/device_config[@revision=$active]/@uuid
        let $instanceid := $doc/../@uuid
        return <item configid="{{$configid}}" instanceid="{{$instanceid}}"/>
        }}</items>
        """
        path = f"{self.root}/{domain}"
        query = query.format(path=path, instance_id=instance_id)

        res = self.dbhandle.query(query)

        return [{"configid": r.attrib["configid"],
                 "instanceid": r.attrib["instanceid"]} for r in
                res.results[0].getchildren()]

    def get_configurations_from_device_name_part(self, domain, device_id_part):
        """
        Returns a list of configurations for a given device
        :param domain: DB domain
        :param device_id_part: part of device name; search is case-insensitive.
        :return: a list of dicts:
            [{"config_id": uuid of the configuration,
              "device_uuid": device instance uuid in the DB,
              "device_id": device instance id},
              ...
            ]
        """
        query = """
        xquery version "3.0";
        let $path := "{path}"
        let $iid := "{instance_id}"
        return <items>{{
        for $doc in collection($path)/xml//device_instance
             [contains(lower-case(@instance_id),lower-case($iid))]
        let $active := $doc/@active_rev
        let $config_id := $doc/device_config[@revision=$active]/@uuid
        let $device_uuid := $doc/../@uuid
        let $device_id := $doc/@instance_id
        return <item config_id="{{$config_id}}"
                device_uuid="{{$device_uuid}}"
                device_id="{{$device_id}}"/>
        }}</items>
        """
        path = f"{self.root}/{domain}"
        query = query.format(path=path, instance_id=device_id_part)

        res = self.dbhandle.query(query)

        return [{"config_id": r.attrib["config_id"],
                 "device_uuid": r.attrib["device_uuid"],
                 "device_id": r.attrib["device_id"]} for r in
                res.results[0].getchildren()]

    def get_projects_from_device(self, domain, uuid):
        """
        Returns the projects which contain a device instance with a given uuid

        :param domain: DB domain
        :param uuid: the uuid of the device instance from the database
        :return: a set containing project names
        """
        query = """
        xquery version "3.0";
        let $path := "{path}"
        let $iid := "{uuid}"
        return <items>{{
        for $doc in collection($path)/xml//device_instance[@uuid=$iid]
        let $serverid := $doc/../../@uuid
        return <item serverid="{{$serverid}}"/>
        }}</items>
        """

        path = f"{self.root}/{domain}"
        query = query.format(path=path, uuid=uuid)
        res = self.dbhandle.query(query)

        servers = [{"serverid": r.attrib["serverid"]} for r in
                   res.results[0].getchildren()]

        query = """
        xquery version "3.0";
        let $path := "{path}"
        let $iid := "{instance}"
        return <items>{{
        for $doc in collection($path)/xml//servers[KRB_Item/uuid=$iid]
        let $projectname := $doc/../../../@simple_name
        return <item projectname="{{$projectname}}"/>
        }}</items>
        """
        projects = set()
        for server in servers:
            queryf = query.format(path=path, instance=server["serverid"])
            res = self.dbhandle.query(queryf)
            projects |= {r.attrib["projectname"] for r in
                         res.results[0].getchildren()}
        return projects

    def get_projects_data_from_device(self, domain, uuid):
        """
        Returns the projects which contain a device instance with a given uuid

        :param domain: DB domain
        :param uuid: the uuid of the device instance from the database
        :return: a list containing project names, uuids and last modification
                 data.
        """
        query = """
        xquery version "3.0";
        let $path := "{path}"
        let $iid := "{uuid}"
        return <items>{{
        for $doc in collection($path)/xml//device_instance[@uuid=$iid]
        let $serverid := $doc/../../@uuid
        return <item serverid="{{$serverid}}"/>
        }}</items>
        """

        path = f"{self.root}/{domain}"
        query = query.format(path=path, uuid=uuid)
        res = self.dbhandle.query(query)

        servers = [{"serverid": r.attrib["serverid"]} for r in
                   res.results[0].getchildren()]

        query = """
        xquery version "3.0";
        declare namespace functx = "http://www.functx.com";
        declare function functx:if-absent(
            $arg as item()* , $value as item()*)  as item()*
        {{ if (exists($arg)) then $arg else $value }};

        let $path := "{path}"
        let $iid := "{instance}"
        return <items>{{
        for $doc in collection($path)/xml//servers[KRB_Item/uuid=$iid]
        let $projectname := $doc/../../../@simple_name
        let $date := functx:if-absent($doc/../../../@date, '')
        let $uuid := $doc/../../../@uuid
        return <item projectname="{{$projectname}}"
                     date="{{$date}}"
                     uuid="{{$uuid}}" />
        }}</items>
        """
        projects = []
        for server in servers:
            queryf = query.format(path=path, instance=server["serverid"])
            res = self.dbhandle.query(queryf)
            for r in res.results[0].getchildren():
                if r.attrib["projectname"] and r.attrib["uuid"]:
                    # Project data is valid; cases of projects with empty
                    # names have been found during tests.
                    projects.append({
                        "projectname": r.attrib["projectname"],
                        "date": r.attrib["date"],
                        "uuid": r.attrib["uuid"]
                    })
        return projects



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from io import StringIO

from lxml import etree

from karabo.common.scenemodel.api import write_scene
from karabo.common.states import State
from karabo.middlelayer import (
    AccessLevel, AccessMode, Device, Hash, Overwrite, Slot, String, TypeHash,
    VectorString, dictToHash, slot)
from karabo.middlelayer.signalslot import Signal
from karabo.native import read_project_model
from karabo.project_db.util import ProjectDBError, get_node, get_project


class ProjectManager(Device):

    # As long as part of Karabo framework, just inherit __version__ from Device

    state = Overwrite(
        defaultValue=State.INIT,
        options=[State.ERROR, State.ON, State.INIT])

    visibility = Overwrite(
        defaultValue=AccessLevel.ADMIN,
        options=[AccessLevel.ADMIN])

    # node containing Database connection details.
    projectDB = get_node()

    domainList = VectorString(
        displayedName="Domain List",
        defaultValue=[],
        description="List of allowed project DB domains. "
                    "Empty list means no restrictions.",
        accessMode=AccessMode.INITONLY)

    signalProjectUpdate = Signal(TypeHash(), String())

    def __init__(self, configuration):
        super().__init__(configuration)
        self.user_db_sessions = {}

    async def onInitialization(self):
        """
        Initialization function of this device. Checks that the configured
        database is indeed reachable and accessible

        If the database is reachable updates the device state to ON,
        otherwise brings the device into ERROR
        """
        try:
            with get_project(self.projectDB, init_db=True):
                self.state = State.ON
        except ProjectDBError as e:
            self.logger.error(f"ProjectDBError : {str(e)}")
            self.state = State.ERROR

    @Slot(displayedName="Reset", allowedStates=[State.ERROR],
          requiredAccessLevel=AccessLevel.EXPERT)
    async def reset(self):
        """Resetting the device brings it into `State.INIT`
        """
        self.state = State.INIT
        await self.onInitialization()

    def _getCurrentConfig(self):
        """
        Return database relevant configuration from expected parameters
        as tuple
        :return: at tuple of host, port
        """
        return self.host.value, self.port.value

    @slot
    def slotGetScene(self, params):
        """Request a scene directly from the database in the correct format

        This protocol is in the style of the capability protocol and expects
        a Hash with parameters:

            name: Name of scene --optional
            domain: Domain to look for the scene item
            uuid: UUID of the scene item
            db_token: db token of the running instance

        :param params: A `Hash` containing the method parameters
        """
        self.logger.debug('Requesting scene directly from database!')

        name = params.get('name', default='')
        # Note: Token used for Karabo >=2.13
        token = params.get('db_token') or params.get('token')
        domain = params.get('domain')
        uuid = [params.get('uuid')]

        # Check if we are already initialized!
        self._checkDbInitialized(token)

        # Start filling the payload Hash
        # ----------------------------------------
        payload = Hash('success', False)
        payload.set('name', name)
        with self.user_db_sessions[token] as db_session:
            try:
                items = db_session.load_item(domain, uuid)
                for item in items:
                    xml = item['xml']
                    item_type = etree.fromstring(xml).get('item_type')
                    if item_type == 'scene':
                        scene = read_project_model(StringIO(xml))
                        payload.set('data', write_scene(scene))
                        payload.set('success', True)
            except ProjectDBError as e:
                self.logger.debug('ProjectDBError in directly loading '
                                  'database scene: {}'.format(e))

        return Hash('type', 'deviceScene',
                    'origin', self.deviceId,
                    'payload', payload)

    @slot
    def slotBeginUserSession(self, token):
        """
        Initialize a DB connection for a user
        :param token: database user token
        """
        self.user_db_sessions[token] = get_project(self.projectDB)
        self.logger.debug("Initialized user session")
        return Hash("success", True)

    @slot
    def slotEndUserSession(self, token):
        """
        End a user session
        :param token: database user token
        """
        del self.user_db_sessions[token]

        self.logger.debug("Ended user session")
        return Hash("success", True)

    def _checkDbInitialized(self, token):
        if token not in self.user_db_sessions:
            raise RuntimeError("You need to init a database session first")

    def _save_items(self, token, items):
        """Internally used method to store items in the project database"""
        savedItems = []
        projectUuids = []
        with self.user_db_sessions[token] as db_session:
            for item in items:
                xml = item.get("xml")
                # Remove XML data to not send it back
                item['xml'] = ''
                uuid = item.get("uuid")
                # XXX: be backward compatible (<2.8.0)!
                item_type = item.get("item_type", "unknown")
                if item_type == "project":
                    projectUuids.append(uuid)
                overwrite = item.get("overwrite")
                domain = item.get("domain")
                exceptionReason = ""
                success = True
                meta = None
                try:
                    meta = db_session.save_item(domain, uuid, xml, overwrite)
                    meta = dictToHash(meta)
                except ProjectDBError as e:
                    success = False
                    exceptionReason = str(e)
                item.set("success", success)
                item.set("reason", exceptionReason)
                item.set("entry", meta)
                savedItems.append(item)

        return savedItems, projectUuids

    @slot
    def slotGenericRequest(self, params):
        """Implements a generic Hash-in/Hash-out interface

        :param params: the input Hash.
            it must contain:
            - a `dbtoken` string matching an open session
            - a `type` string matching a non-generic slot name
            - the optional fields required by the non-generic slot
            requested.
        """
        try:
            msg = "Input must be a Hash"
            assert isinstance(params, Hash), msg
            msg = "'type' must be present in the input hash"
            assert "type" in params, msg
            msg = "'token' must be present in the input hash"
            assert "token" in params, msg

            action_type = params["type"]
            token = params["token"]  # session token
            if action_type == "listItems":
                return self.slotListItems(token,
                                          params['domain'],
                                          params.get('item_types', None))
            elif action_type == "loadItems":
                return self.slotLoadItems(token, params['items'])
            elif action_type == "listDomains":
                return self.slotListDomains(token)
            elif action_type == "updateAttribute":
                return self.slotUpdateAttribute(token, params['items'])
            elif action_type == "saveItems":
                return self.slotSaveItems(token, params['items'],
                                          params.get('client', None))
            elif action_type == "beginUserSession":
                return self.slotBeginUserSession(token)
            elif action_type == "endUserSession":
                return self.slotEndUserSession(token)
            elif action_type == "listProjectsWithDevice":
                return self.slotListProjectsWithDevice(params)
            else:
                raise NotImplementedError(f"{type} not implemented")
        except Exception as e:
            return Hash("success", False, "reason", f"{type(e)}: {e}")

    @slot
    def slotSaveItems(self, token, items, client=None):
        """Save items in project database

        :param token: database user token
        :param items: items to be save. Should be a list(Hash) object were each
            entry is of the form:

            - xml: xml of item
            - uuid: uuid of item
            - overwrite: behavior in case of conflict
            - domain: to write to
        :param client: the client information (string) if provided

        :raises: `ProjectDBError` in case of database (connection) problems
            `TypeError` in case an no type information or an unknown type
            is found in the root element.
            `RuntimeError` if no database if connected.
        """

        self.logger.debug("Saving items: {}".format([i.get("uuid") for i in
                                                     items]))
        self._checkDbInitialized(token)
        saved, uuids = self._save_items(token, items)
        if client and uuids:
            self.signalProjectUpdate(Hash("projects", uuids,
                                          "client", client),
                                     self.deviceId)

        return Hash('items', saved)

    @slot
    def slotLoadItems(self, token, items):
        """
        Loads items from the database

        :param token: database user token
        :param items: list of Hashes containing information on which items
            to load. Each list entry should be a Hash containing
            - uuid: the uuid of the item
            - domain: domain to load item from

        :return: a Hash where the keys are the item uuids and values are the
            item XML. If the load failed the value for this uuid is set
            to False
        """

        self.logger.debug("Loading items: {}"
                          .format([i.get("uuid") for i in items]))

        self._checkDbInitialized(token)

        exceptionReason = ""
        success = True
        loadedItems = []
        with self.user_db_sessions[token] as db_session:
            # verify that items belong to single domain
            domain = items[0].get("domain")
            keys = [it.get('uuid') for it in items
                    if it.get('domain') == domain]
            assert len(keys) == len(items), "Incorrect domain given!"

            try:
                items = db_session.load_item(domain, keys)
                for item in items:
                    uuid = item["uuid"]
                    h = Hash("domain", domain,
                             "uuid", uuid,
                             "xml", item["xml"])
                    loadedItems.append(h)
                    # Remove from the list of requested keys
                    keys.remove(uuid)

                # Any keys left were not in the database
                if len(keys) > 0:
                    success = False
                    exceptionReason = f'Items "{keys}" not found!'
            except ProjectDBError as e:
                exceptionReason = str(e)
                success = False
        return Hash('items', loadedItems,
                    'success', success,
                    'reason', exceptionReason)

    @slot
    def slotListItems(self, token, domain, item_types=None):
        """
        List items in domain which match item_types if given, or all items
        if not given

        :param token: database user token
        :param domain: domain to list items from
        :param item_types: list or tuple of item_types to list
        :return: a list of Hashes where each entry has keys: uuid, item_type
            and simple_name
        """
        self._checkDbInitialized(token)

        with self.user_db_sessions[token] as db_session:
            exceptionReason = ""
            success = True
            resHashes = []
            try:
                res = db_session.list_items(domain, item_types)
                for r in res:
                    h = Hash('uuid', r['uuid'],
                             'item_type', r['item_type'],
                             'simple_name', r['simple_name'],
                             'is_trashed', r['is_trashed'],
                             'date', r['date'],
                             'user', r['user'])
                    h.set('description', r['description'])
                    resHashes.append(h)
            except ProjectDBError as e:
                exceptionReason = str(e)
                success = False
        return Hash('items', resHashes,
                    'success', success,
                    'reason', exceptionReason)

    @slot
    def slotListNamedItems(self, token, domain, item_type, simple_name):
        """
        List items in domain which match item_type and simple_name

        :param token: database user token
        :param domain: domain to list items from
        :param item_type: item_type to match
        :param simple_name: simple_name to match
        :return: a list of Hashes where each entry has keys: uuid, date,
            item_type and simple_name sorted by date
        """
        self._checkDbInitialized(token)

        with self.user_db_sessions[token] as db_session:
            exceptionReason = ""
            success = True
            resHashes = []
            try:
                res = db_session.list_named_items(
                    domain, item_type, simple_name)
                for r in res:
                    h = Hash('uuid', r['uuid'],
                             'item_type', r['item_type'],
                             'simple_name', r['simple_name'],
                             'is_trashed', r['is_trashed'],
                             'date', r['date'],
                             'user', r['user'])
                    h.set('description', r['description'])
                    resHashes.append(h)
                resHashes.sort(key=lambda x: x['date'])
            except ProjectDBError as e:
                exceptionReason = str(e)
                success = False
        return Hash('items', resHashes,
                    'success', success,
                    'reason', exceptionReason)

    @slot
    def slotListDomains(self, token):
        """
        List domains available on this database

        :param token: database user token
        :return:
        """
        self._checkDbInitialized(token)

        with self.user_db_sessions[token] as db_session:
            success = True
            exceptionReason = ""
            res = []
            try:
                res = db_session.list_domains()
                if len(self.domainList):
                    res = [domain for domain in res
                           if domain in self.domainList]
            except Exception as e:
                exceptionReason = str(e)
                success = False
            return Hash('success', success,
                        'domains', res,
                        'reason', exceptionReason)

    @slot
    def slotListProjectsWithDevice(self, args):
        """
        List projects in domain which have configurations for a given device.

        :param args: a hash that must contain the keys "token", "domain" and
            "device_id" with the following meanings: "token" is the
            database user token, "domain" is the domain to list
            projects from and "device_id" is the part that must be
            contained in the ids of the devices with configurations
            stored in the projects to be listed.

        :return: a Hash with key, "projects", with a list of Hashes for its
            value. Each Hash in the list has four keys: "uuid",
            "name", "last_modified" and "devices". "uuid" is the unique id
            of the project in the project database, "name" is the project
            name, "last_modified" is the UTC imestamp of the projects's
            most recent modification in "%Y-%m-%d %H:%M:%S" format and
            "devices" is a list of the device ids in the project that
            match the given device_id part passed to the slot.
            The returned Hash also has a boolean key "success" that
            indicates whether the slot execution has been successful
            (True) and a string key "reason", that will contain an error
            description when the slot execution fails.
        """
        for k in ["token", "domain", "device_id"]:
            if not args.has(k):
                return Hash('success', False,
                            'reason', f'Key "{k}" missing in "args" hash.')

        token = args["token"]
        domain = args["domain"]
        device_id = args["device_id"]

        self._checkDbInitialized(token)

        with self.user_db_sessions[token] as db_session:
            exceptionReason = ""
            success = True
            try:
                res_prjs = []
                prjs = db_session.get_projects_with_device(domain, device_id)
                for prj in prjs:
                    res_prjs.append(
                        Hash("name", prj["projectname"],
                             "uuid", prj["uuid"],
                             "last_modified", prj["date"],
                             "devices", prj["devices"]))
            except ProjectDBError as e:
                exceptionReason = str(e)
                success = False
            return Hash('projects', res_prjs,
                        'success', success,
                        'reason', exceptionReason)

    @slot
    def slotUpdateAttribute(self, token, items):
        """
        Update any attribute of given ``items`` in the database

        :param token: database user token
        :param items: list of Hashes containing information on which items
                      to update. Each list entry should be a Hash containing

                      - domain: domain the item resides at
                      - uuid: the uuid of the item
                      - item_type: indicate type of item which attribute should
                                   be changed
                      - attr_name: name of attribute which should be changed
                      - attr_value: value of attribute which should be changed

        :return: a list of Hashes where each entry has keys: domain, item_type,
                 uuid, attr_name, attr_value
        """

        self._checkDbInitialized(token)

        with self.user_db_sessions[token] as db_session:
            exceptionReason = ""
            success = True
            resHashes = []
            try:
                res = db_session.update_attributes(items)
                for r in res:
                    h = Hash('domain', r['domain'],
                             'item_type', r['item_type'],
                             'uuid', r['uuid'],
                             'attr_name', r['attr_name'],
                             'attr_value', r['attr_value'])
                    resHashes.append(h)
            except ProjectDBError as e:
                exceptionReason = str(e)
                success = False
            return Hash('items', resHashes,
                        'success', success,
                        'reason', exceptionReason)

    @slot
    def slotListProjectAndConfForDevice(self, token, domain, deviceId):
        self._checkDbInitialized(token)

        with self.user_db_sessions[token] as db_session:
            exceptionReason = ""
            success = True
            resHashes = []
            try:
                res = db_session.get_projects_with_conf(domain, deviceId)
                resHashes = [Hash("project_name", k, "active_config_ref", v)
                             for k, v in res.items()]
            except ProjectDBError as e:
                exceptionReason = str(e)
                success = False
            return Hash('items', resHashes,
                        'success', success,
                        'reason', exceptionReason)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
#
# Model classes generated from the existing Sqlite database by pwiz,
# a model generator script bundled with peewee. The command to generate
# the model is:
#
#     python -m pwiz -e sqlite ${path_to_sqlite_db}
#
# All the STRING(max_size) fields in the database were generated by pwiz
# as UnknownField instances. This wouldn't work, so they were manually
# changed to be of type peewee.CharField(max_length=max_size) and the
# UnknownField class, which was only an empty bodied class generated by
# pwiz to work as a placeholder, has been removed.
#
# An auxiliary model class generated by pwiz, SqliteSequence, wasn't actually
# being used (probably required by some older peewee version?) and has been
# removed.
#
# pwiz also generates a BaseModel class that binds the models statically to
# the database connection. That was not suitable for the ConfigurationManager
# and has been replaced by a dynamic binding to the database from within
# 'ConfigurationManager.dbHandle.__enter__(self)'.
#
import datetime

from peewee import (
    SQL, BooleanField, CharField, Check, DateTimeField, ForeignKeyField,
    IntegerField, Model, TextField)


class ConfigSchema(Model):
    digest = CharField(max_length=40, unique=True)
    schema_data = TextField()

    class Meta:
        table_name = 'ConfigSchema'


class ConfigSet(Model):
    config_name = CharField(max_length=80, index=True)
    description = TextField(null=True)
    device_set_id = CharField(max_length=40, index=True)
    overwritable = BooleanField(constraints=[SQL("DEFAULT FALSE")])
    priority = IntegerField(constraints=[SQL("DEFAULT 1")],
                            index=True, null=True)
    user = CharField(max_length=64,
                     constraints=[SQL("DEFAULT '.'")])

    class Meta:
        table_name = 'ConfigSet'
        indexes = (
            (('device_set_id', 'config_name'), True),
        )
        constraints = [
            Check('priority > 0 AND priority < 4'),
        ]


class DeviceConfig(Model):
    config_data = TextField()
    config_set = ForeignKeyField(column_name='config_set_id',
                                 field='id', model=ConfigSet)
    device_id = CharField(max_length=1024, index=True)
    schema = ForeignKeyField(column_name='schema_id',
                             field='id', model=ConfigSchema)
    timestamp = DateTimeField(default=datetime.datetime.now)

    class Meta:
        table_name = 'DeviceConfig'
        indexes = (
            (('config_set', 'device_id'), True),
        )



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import hashlib
import os
from contextlib import ContextDecorator
from datetime import datetime

from peewee import DoesNotExist, SqliteDatabase, fn

from .models import ConfigSchema, ConfigSet, DeviceConfig
from .utils import (
    CONFIG_DB_CONFIG_SET_ID, CONFIG_DB_DATA, CONFIG_DB_DESCRIPTION,
    CONFIG_DB_DEVICE_ID, CONFIG_DB_DIFF_TIMEPOINT, CONFIG_DB_MAX_TIMEPOINT,
    CONFIG_DB_MIN_TIMEPOINT, CONFIG_DB_NAME, CONFIG_DB_OVERWRITABLE,
    CONFIG_DB_PRIORITY, CONFIG_DB_SCHEMA, CONFIG_DB_TIMEPOINT, CONFIG_DB_USER,
    ISO8601_FORMAT, ConfigurationDBError, create_config_set_id,
    datetime_from_string)

CONFIGURATION_LIMIT = 300


class DbHandle(ContextDecorator):
    """The DbHandle class create a configuration database connection

    XXX: Eventual authorization is handled here...
    """

    def __init__(self, dbName):
        self.dbName = dbName
        self.dbConn = SqliteDatabase(
            self.dbName,
            pragmas={
                'journal_mode': 'wal',
                'foreign_key': 1,
                'ignore_check_constraints': 0
            }
        )

    def __enter__(self):
        self.dbConn.bind(
            [ConfigSet, ConfigSchema, DeviceConfig]
        )
        self.dbConn.connect(reuse_if_open=True)
        return self.dbConn

    def __exit__(self, type, value, traceback):
        self.dbConn.close()

    def authenticate(self):
        return True


class ConfigurationDatabase:

    def __init__(self, db):
        self.dbHandle = db

    @property
    def path(self):
        return self.dbHandle.dbName

    @property
    def dirname(self):
        return os.path.dirname(self.dbHandle.dbName)

    def assureExisting(self):
        with self.dbHandle as db:
            db.create_tables(
                [ConfigSchema, ConfigSet, DeviceConfig],
                safe=True
            )

    # Public Interface
    # --------------------------------------------------------------------

    # TODO: ponder with others and implement, depending on the results of the
    # pondering process, methods to retrieve all devices, all config sets,
    # and all the number of saved configs for all devices in the Db. Those
    # methods would support a ConfigDb Manager application. Also ponder about
    # adding the most recent timestamp when a configuration has been applied.
    # That information might be useful for someone with a ConfigDb admin role
    # that wants to get rid of stale configurations. Also add the corresponding
    # method to delete a configuration - only accessible to users with admin
    # role (users might be real people or scheduled maintenance jobs.

    def list_configurations(self, deviceId, name_part=''):
        """Retrieves the set of device configurations related to a name part

        :param deviceId: the id of the device
        :param name_part: the name part of the device configurations. empty
                          means all the named configurations.

        :returns: list of configuration records (can be empty). Each configu-
                  ration record is a dictionary. The value for the key
                  'config_set_id' os either -1, if the configuration is the
                  only that has been saved as part of the save operation, or it
                  is the id of the config_set (and its value can be used to
                  retrieve the remaining configs that were saved by the same
                  save operation).
        """
        with self.dbHandle:
            cfgs = (
                DeviceConfig
                .select(
                    ConfigSet.config_name, DeviceConfig.timestamp,
                    ConfigSet.description, ConfigSet.priority,
                    ConfigSet.user, ConfigSet.overwritable,
                    ConfigSet.device_set_id, ConfigSet.id
                )
                .join(ConfigSet)
            )
            if name_part:
                # When a name_part is given, filter configs by it.
                cfgs = cfgs.where(
                    DeviceConfig.device_id == deviceId,
                    ConfigSet.config_name.contains(name_part)
                )
            else:
                # Otherwise, only filters by device_id
                cfgs = cfgs.where(
                    DeviceConfig.device_id == deviceId
                )
            cfgs = cfgs.tuples()

            ret_recordings = []
            device_set_digest = create_config_set_id([deviceId])
            for cfg in cfgs:
                conf_dict = {}
                conf_dict[CONFIG_DB_NAME] = cfg[0]
                conf_dict[CONFIG_DB_TIMEPOINT] = (
                    cfg[1].strftime(ISO8601_FORMAT))
                conf_dict[CONFIG_DB_DESCRIPTION] = cfg[2]
                conf_dict[CONFIG_DB_PRIORITY] = cfg[3]
                conf_dict[CONFIG_DB_USER] = cfg[4]
                conf_dict[CONFIG_DB_OVERWRITABLE] = cfg[5]

                # If the config isn't the only config in its device set,
                # returns the config_set_id (the prim. key of the device set)
                # for the key CONFIG_DB_CONFIG_SET_ID; Otherwise returns -1 to
                # flag that the config is the only one saved in its device set.
                if device_set_digest == cfg[6]:
                    # the id generated for the device_set based on the
                    # device_id matches the digest in the database;
                    # that means the device configuration is the only one
                    # that has been saved as part of the config_set save.
                    conf_dict[CONFIG_DB_CONFIG_SET_ID] = -1
                else:
                    conf_dict[CONFIG_DB_CONFIG_SET_ID] = cfg[7]

                ret_recordings.append(conf_dict)

            return ret_recordings

    def list_configurations_in_set(self, config_set_id):
        """Retrieves the set of configurations in a given config set.

        :param config_set_id: the id of the config set

        :returns: list of configuration records (one or more). Each configu-
                  ration record is a dictionary with the keys:
                      "deviceId",
                      "config",
                      "schema",
                      "timepoint".
                  the list of configurations is sorted by "deviceId".
        """
        with self.dbHandle:
            cfgs = (
                DeviceConfig
                .select(
                    DeviceConfig.device_id, DeviceConfig.config_data,
                    ConfigSchema.schema_data, DeviceConfig.timestamp,
                )
                .join(ConfigSchema)
                .where(
                    DeviceConfig.config_set_id == config_set_id
                )
                .order_by(DeviceConfig.device_id)
            )
            cfgs = cfgs.tuples()

            ret_recordings = []
            for cfg in cfgs:
                conf_dict = {}
                conf_dict[CONFIG_DB_DEVICE_ID] = cfg[0]
                conf_dict[CONFIG_DB_DATA] = cfg[1]
                conf_dict[CONFIG_DB_SCHEMA] = cfg[2]
                conf_dict[CONFIG_DB_TIMEPOINT] = (
                    cfg[3].strftime(ISO8601_FORMAT))
                ret_recordings.append(conf_dict)

            return ret_recordings

    def list_devices(self, priority=3):
        """Retrieves the set of deviceIds related to configuration priority

        :param priority: the desired priority

        :returns: list of deviceIds records (can be empty).
        """
        with self.dbHandle:
            devs = (
                DeviceConfig
                .select(DeviceConfig.device_id.distinct())
                .join(ConfigSet)
                .where(ConfigSet.priority == priority)
                .tuples()
            )
            ret = [dev[0] for dev in devs]
            return ret

    # TODO: min_timepoint and max_timepoint have lost their precision since
    #       the introduction of minSetSize: the gathering of configs in the
    #       config set is not exhaustive for performance reasons and that
    #       means that it is not guaranteed that all timestamps for all
    #       configs in the set have been retrieved. The proposed solution is
    #       to remove MIN_TIMEPOINT, MAX_TIMEPOINT and DIFF_TIMEPOINT from
    #       the response records and add REF_TIMEPOINT which should be the
    #       average timepoint of the timepoints actually retrieved.
    def list_configuration_sets(self, deviceIds, minSetSize=None):
        """Retrieves the sets of configurations related to a list of devices.
        Sets containing configurations for either a subset or a superset of the
        list of devices can be returned depending on the value of the
        parameter minSetSize.

        The goal of returning config sets that do not have all the devices in
        the list (or that have more than the devices in the list) is to support
        configuration sets for components (sets of devices) whose composition
        changes through their lifetime.

        If no minSetSize is specified, all configuration sets containing at
        least the deviceIds specified will be retrieved (supersets). On the
        other hand, a minSetSize lower than the number of devices specified
        will allow retrieving subsets of the devices set.

        :param deviceIds: list of deviceIds whose configurations should be
                          listed.

        :param minSetSize: minimum size of configuration sets to retrieve. If
                           no value or a value lower than 1 is specified,
                           len(deviceIds) is used as minSetSize. If a value
                           greater than the number of devices is specified,
                           an exception of type ConfigurationDBError is
                           thrown.

        :returns: list of configuration records (can be empty). Each configu-
                  ration record is a dictionary. The configurations timestamps
                  are returned in ISO8601 format (yyyy-mm-ddThh:mm:ss.micro).

        Example: Retrieve configuration sets for devices DevA, DevB and DevC.
                 Let's assume that in the database, DevA has configurations
                 named {Nam1, Nam2, Nam3}, DevB has configurations named
                 {Nam2, Nam3} and DevC has configurations named
                 {Nam1, Nam3}. If minSetSize is not specified (or is lower
                 than one), it will assume the value 3 (the number of
                 deviceIds given). Only one set will be retrieved, as there is
                 only one set with configurations for all the given devices,
                 the set with the configuration named Nam3. If a minSetSize of
                 2 is specified, three sets will be returned: one for the two
                 configurations Nam1 saved for DevA and DevC, another for the
                 two configurations Nam2 saved for DevA and DevB and the same
                 set of size 3 returned when minSetSize was not specified.

        """
        if len(deviceIds) == 0:
            raise ConfigurationDBError('Please provide at least one device id')

        if not minSetSize or minSetSize < 0:
            minSetSize = len(deviceIds)
        else:
            if minSetSize > len(deviceIds):
                raise ConfigurationDBError(
                    'minSetSize has to be between 1 and the number of '
                    f'deviceIds, {len(deviceIds)}.')

        with self.dbHandle:
            DevCfgAlias = DeviceConfig.alias()
            query = (
                DevCfgAlias
                .select(
                    fn.COUNT(DevCfgAlias.device_id).alias('num_cfgs'),
                    ConfigSet.id.alias('config_set_id'), ConfigSet.config_name,
                    ConfigSet.description, ConfigSet.priority, ConfigSet.user,
                    ConfigSet.overwritable,
                    fn.MIN(DevCfgAlias.timestamp).alias('min_timepoint'),
                    fn.MAX(DevCfgAlias.timestamp).alias('max_timepoint'),
                )
                .join(ConfigSet)
                .where(DevCfgAlias.device_id << deviceIds)
                .group_by(ConfigSet.config_name, ConfigSet.device_set_id)
            )
            cfgs = (
                DeviceConfig
                .select(
                    query.c.num_cfgs, query.c.config_name,
                    query.c.description, query.c.priority,
                    query.c.user, query.c.overwritable,
                    query.c.min_timepoint, query.c.max_timepoint,
                    query.c.config_set_id
                )
                .from_(query)
                .where(query.c.num_cfgs >= minSetSize)
                .tuples()
            )

            ret_recordings = []
            for cfg in cfgs:
                conf_dict = {}
                conf_dict[CONFIG_DB_NAME] = cfg[1]
                conf_dict[CONFIG_DB_DESCRIPTION] = cfg[2]
                conf_dict[CONFIG_DB_PRIORITY] = cfg[3]
                conf_dict[CONFIG_DB_USER] = cfg[4]
                conf_dict[CONFIG_DB_OVERWRITABLE] = cfg[5]
                # NOTE: both timepoints will be retrieved as strings. SQLite
                # doesn't have a real datetime type; as there was no explicit
                # Model.DateTimeField involved, the peewee fabricated query
                # will return a string. This may change when a database that
                # has a real datetime type (e.g. Postgres) is used.
                min_time = datetime_from_string(cfg[6])
                max_time = datetime_from_string(cfg[7])
                conf_dict[CONFIG_DB_MIN_TIMEPOINT] = (
                    min_time.strftime(ISO8601_FORMAT))
                conf_dict[CONFIG_DB_MAX_TIMEPOINT] = (
                    max_time.strftime(ISO8601_FORMAT))
                diff_time = max_time - min_time
                conf_dict[CONFIG_DB_DIFF_TIMEPOINT] = diff_time.total_seconds()
                # Differently from list_configurations, where -1 is returned
                # for ConfigSets with only one configuration,
                # list_configurations_sets always returns the id of the
                # ConfigSet.
                conf_dict[CONFIG_DB_CONFIG_SET_ID] = cfg[8]

                ret_recordings.append(conf_dict)

            return ret_recordings

    def check_configurations_limits(self, deviceIds):
        """Checks if any of the devices exceeds the configuration limit.

        :param deviceIds: list of deviceIds who should be validated

        :returns: boolean (True or False)
        """
        with self.dbHandle:
            return self._check_configurations_limits(deviceIds)

    def get_configuration(self, deviceId, name):
        """Retrieves a device configuration given its name.

        :param deviceId: the id of the device.
        :param name: the name of the device configuration

        :returns: dictionary result (can be empty). The configuration
                  timestamp is returned in ISO8601 format
                  (yyyy-mm-ddThh:mm:ss.micro).
        """
        with self.dbHandle:
            cfg = (
                DeviceConfig
                .select(
                    ConfigSet.config_name, DeviceConfig.timestamp,
                    DeviceConfig.config_data, ConfigSchema.schema_data,
                    ConfigSet.description, ConfigSet.priority,
                    ConfigSet.user, ConfigSet.overwritable
                )
                .join(
                    ConfigSet,
                    on=(DeviceConfig.config_set == ConfigSet.id))
                .join(
                    ConfigSchema,
                    on=(DeviceConfig.schema == ConfigSchema.id))
                .where(
                    DeviceConfig.device_id == deviceId,
                    ConfigSet.config_name == name
                )
                .tuples()
            )
            return self._config_to_dict(cfg[0]) if len(cfg) > 0 else {}

    def get_last_configuration(self, deviceId, priority=1):
        """Retrieves the most recent device configuration with a priority.

        :param deviceId: the id of the device.
        :param priority: priority associated to the configuration to return.

        :returns: dictionary result (can be empty). The configuration timestamp
                  is returned in ISO8601 format (yyyy-mm-ddThh:mm:ss.micro).
        """
        with self.dbHandle:
            cfg = (
                DeviceConfig
                .select(
                    ConfigSet.config_name, DeviceConfig.timestamp,
                    DeviceConfig.config_data, ConfigSchema.schema_data,
                    ConfigSet.description, ConfigSet.priority,
                    ConfigSet.user, ConfigSet.overwritable
                )
                .join(
                    ConfigSet,
                    on=(DeviceConfig.config_set == ConfigSet.id))
                .join(
                    ConfigSchema,
                    on=(DeviceConfig.schema == ConfigSchema.id))
                .where(
                    DeviceConfig.device_id == deviceId,
                    ConfigSet.priority == priority
                )
                .order_by(DeviceConfig.timestamp.desc())
                .tuples()
            )

            return self._config_to_dict(cfg[0]) if len(cfg) > 0 else {}

    def save_configuration(self, name, configs, description="", user=".",
                           priority=1, timestamp="", overwritable=False):
        """Saves one or more device configurations (and schemas).

        The configs and schemas are saved for multiple devices and timepoints
        in one roundtrip to the database. They share a common name,
        description, priority and user.

        :param name: the configuration(s) name
        :param configs: a list of dictionaries with the following keys:
                        - 'deviceId',
                        - 'config'
                        - 'schema'

                Here, 'config' and 'schema' are expected to be
                strings with the Base64 encoded form of the binary
                serialization of a device config and its corresponding schema.

        :param description: optional description for the named configuration.

        :param user: The default value, '.', means anonymous.

        :param priority: Must be an integer between 1 and 3.

        :param timestamp: The timestamp to be associated with the device
            configuration. If no timestamp is given, the current UTC timestamp
            with milliseconds precision is used. The timestamp should be UTC
            and specified as a string in ISO8601 format.

        :param overwritable: Will the configuration be overwritable?
            Ignored if the named configuration already exists in the database
            for the set of devices in configs.
        """
        if not 1 <= priority <= 3:
            raise ConfigurationDBError(
                "Please provide a priority value between 1 and 3.")

        if len(name) >= 80:  # 80 is the size of the underlying db field
            raise ConfigurationDBError(
                "Please provide a name with less than 80 characters.")

        reqKeys = [CONFIG_DB_DEVICE_ID, CONFIG_DB_DATA, CONFIG_DB_SCHEMA]
        for index, config in enumerate(configs):
            for key in reqKeys:
                if key not in config.keys():
                    raise ConfigurationDBError(
                        f'Configuration #{index} is missing key "{key}".')

        deviceIds = [conf[CONFIG_DB_DEVICE_ID] for conf in configs]

        deviceSetId = create_config_set_id(deviceIds)

        with self.dbHandle as db:

            if self._is_config_name_taken(name, deviceIds, deviceSetId):
                raise ConfigurationDBError(
                    f"The config name {name} is already "
                    f"taken for at least one of the device(s) {deviceIds}")

            overwrite_id = self._get_overwrite_id(name, deviceSetId)
            with db.atomic() as txn:
                try:
                    if overwrite_id:
                        # An overwritable configuration with the name exists
                        # for the set of devices; update it.
                        self._update_configs(
                            overwrite_id, configs, description,
                            user, int(priority), timestamp=timestamp)
                    else:
                        # No configuration with the name exists for the set of
                        # devices; Add it after checking that the limit of
                        # configurations per device stored in the database will
                        # be respected.
                        is_limit = self._check_configurations_limits(deviceIds)
                        if is_limit:
                            raise ConfigurationDBError(
                                f'One of the devices "{deviceIds}" would '
                                f'exceed the number  of configurations per '
                                f'device, {CONFIGURATION_LIMIT}.')
                        self._insert_configs(
                            deviceSetId, name, configs, description,
                            user, int(priority), timestamp=timestamp,
                            overwritable=overwritable)
                except Exception as e:
                    txn.rollback()
                    raise ConfigurationDBError(f"{e}")
                else:
                    txn.commit()  # An explicit commit is not really needed.

    def is_config_name_taken(self, name, deviceIds, deviceSetId=None):
        """Checks if a config name is already in use by any of the devices in a
        list.

        :param  name: the configuration name to be checked.
        :param deviceIds: list of deviceIds whose taken configurations names
                          should be searched.
        :param deviceSetId: identifier for the set of devices in deviceIds.


        :returns: True if the config name is already taken by any of the
                  devices in the list.
        """
        with self.dbHandle:
            return self._is_config_name_taken(name, deviceIds, deviceSetId)

    # Private Interface
    # --------------------------------------------------------------------

    def _is_config_name_taken(self, name, deviceIds, deviceSetId=None):
        """The actual core of the public is_config_name_taken. The public
        wrapper takes care of creating a connection context. The private
        core uses the connection context from its caller.
        """
        ro_cfg_with_name = (
            DeviceConfig
            .select(fn.COUNT(DeviceConfig.id))
            .join(ConfigSet)
            .where(
                ConfigSet.config_name == name,
                ConfigSet.overwritable == False,  # noqa: E712
                DeviceConfig.device_id << deviceIds
            )
            .scalar()
        )
        if ro_cfg_with_name > 0:
            # If there's already any read-only config with the same
            # name for any of the devices, the name is taken.
            return True

        if not deviceSetId:
            deviceSetId = create_config_set_id(deviceIds)

        ovwr_cfgs = (
            ConfigSet
            .select(ConfigSet.device_set_id)
            .where(
                ConfigSet.config_name == name,
                ConfigSet.overwritable == True  # noqa: E712
            )
            .tuples()
        )
        if (
            len(ovwr_cfgs) == 0 or
            (len(ovwr_cfgs) == 1 and ovwr_cfgs[0][0] == deviceSetId)
        ):
            # The name is not taken if there's no other overwritable
            # config with the same name or if there's only one and it is
            # for the same device set.
            return False
        else:
            return True

    def _check_configurations_limits(self, deviceIds):
        """The actual core of the public _check_configuration_limits.
        The public wrapper takes care of creating a connection context. The
        private core uses the connection context from its caller.
        """
        DevConfigAlias = DeviceConfig.alias()
        cnt_qry = (
            DevConfigAlias
            .select(
                DevConfigAlias.device_id,
                fn.COUNT(DevConfigAlias.config_data).alias('num_cfgs')
            )
            .where(DevConfigAlias.device_id << deviceIds)
            .group_by(DevConfigAlias.device_id)
        )
        devs_limit = (
            DeviceConfig
            .select(fn.COUNT(cnt_qry.c.device_id))
            .from_(cnt_qry)
            .where(cnt_qry.c.num_cfgs >= CONFIGURATION_LIMIT)
            .scalar()
        )
        return devs_limit > 0

    def _schema_id_from_digest(self, schemaDigest):
        """Gets the schema id for a given digest.

        :return: -1 if there's no schema with the given digest in the db;
                 otherwise returns the id of the schema in the database.
        """
        try:
            sch_id = (
                ConfigSchema.get(ConfigSchema.digest == schemaDigest).id
            )
            return sch_id
        except DoesNotExist:
            return -1

    def _get_overwrite_id(self, name, set_id):
        """Gets an overwritable configuration for a given set of devices
           from the database.

        :param name: the name of the overwritable configuration.
        :param set_id: the set identifier (unique hash for each set of devices)
        :return: the internal identifier of the overwritable config set, or
                 None if no config set has been found.
        """
        try:
            overwrite_id = (
                ConfigSet
                .select(ConfigSet.id)
                .where(
                    ConfigSet.device_set_id == set_id,
                    ConfigSet.config_name == name
                )
                .get()
            )
            return overwrite_id
        except DoesNotExist:
            return None

    def _get_config_set(self, name, set_id):
        """Gets the config set id for a given name and set digest.

        :param name: the name of the configuration.
        :param set_id: the set identifier (unique hash for each set of devices)
        :return: the internal identifier of the config set, or None if no
                 config set has been found.
        """
        try:
            cfg_set_id = (
                ConfigSet.get(
                    ConfigSet.device_set_id == set_id,
                    ConfigSet.config_name == name
                ).id
            )
            return cfg_set_id
        except DoesNotExist:
            return None

    def _config_to_dict(self, cfg):
        """Private method to transform a config fecthed from the database into
           a dictionary.
        """
        conf_dict = {}
        if cfg:
            conf_dict[CONFIG_DB_NAME] = cfg[0]
            conf_dict[CONFIG_DB_TIMEPOINT] = (
                cfg[1].strftime(ISO8601_FORMAT))
            conf_dict[CONFIG_DB_DATA] = cfg[2]
            conf_dict[CONFIG_DB_SCHEMA] = cfg[3]
            conf_dict[CONFIG_DB_DESCRIPTION] = cfg[4]
            conf_dict[CONFIG_DB_PRIORITY] = cfg[5]
            conf_dict[CONFIG_DB_USER] = cfg[6]
            conf_dict[CONFIG_DB_OVERWRITABLE] = cfg[7]

        return conf_dict

    def _update_configs(self, set_id, configs, description, user,
                        priority, timestamp):
        """Updates overwritable device configurations in the database after
           updating the corresponding device set record."""
        # Starts the transaction and writes each of the configs,
        # committing at the end or rolling back if any error happens.
        ConfigSet.update(
            user=user, priority=priority, description=description
        ).where(ConfigSet.id == set_id).execute()
        for config in configs:
            self._update_config(set_id, config, timestamp)

    def _update_config(self, set_id, config, timestamp):
        """Updates a device configuration in the database.

        NOTE: It's up to the caller to guarantee that the set_id refers to an
              overwritable configuration set.
        """
        schema_id = self._insert_schema_if_needed(config[CONFIG_DB_SCHEMA])

        if len(timestamp) > 0:
            timestamp_val = datetime_from_string(timestamp)
        else:
            timestamp_val = datetime.now()

        update_count = DeviceConfig.update(
            config_data=config[CONFIG_DB_DATA], schema_id=schema_id,
            timestamp=timestamp_val
        ).where(
            DeviceConfig.config_set_id == set_id,
            DeviceConfig.device_id == config[CONFIG_DB_DEVICE_ID]
        ).execute()

        if update_count == 0:
            # There was no device config record to update. This may happen
            # for overwritable configurations that are saved with injected
            # device_set_id values. During the first multi chunk save of
            # such configurations, any save operation that is not for the
            # first chunk of device configs to be saved may not trigger any
            # update during the save.
            self._insert_config(set_id, config, timestamp)

    def _insert_configs(self, set_digest, name, configs, description,
                        user, priority, timestamp, overwritable):
        """Saves new device configurations in the database after saving the
           corresponding device set record."""
        # Starts the transaction and writes each of the configs,
        # committing at the end or rolling back if any error happens.
        set_id = self._get_config_set(name, set_digest)
        if not set_id:
            # There's no config set yet for the configurations to be saved.
            # Insert it.
            set_id = ConfigSet.insert(
                device_set_id=set_digest, config_name=name,
                user=user, priority=priority, description=description,
                overwritable=overwritable
            ).execute()
        for config in configs:
            self._insert_config(set_id, config, timestamp)

    def _insert_config(self, set_id, config, timestamp):
        """Saves a new device configuration in the database."""
        schema_id = self._insert_schema_if_needed(config[CONFIG_DB_SCHEMA])
        if len(timestamp) > 0:
            cfg_id = DeviceConfig.insert(
                config_set_id=set_id, device_id=config[CONFIG_DB_DEVICE_ID],
                timestamp=datetime_from_string(timestamp),
                config_data=config[CONFIG_DB_DATA], schema_id=schema_id
            ).execute()
        else:
            cfg_id = DeviceConfig.insert(
                config_set_id=set_id, device_id=config[CONFIG_DB_DEVICE_ID],
                config_data=config[CONFIG_DB_DATA], schema_id=schema_id
            ).execute()
        return cfg_id

    def _insert_schema_if_needed(self, sch_data):
        """Inserts an schema if it's digest reveals it's not stored yet.

           :param db: handle to an open database connection.
           :param sch_data: serialized form of the schema to store.
           :return: id of the existing or just created schema record.
        """
        sch_digest = hashlib.sha1(sch_data.encode('utf-8')).hexdigest()
        sch_id = self._schema_id_from_digest(sch_digest)
        if sch_id == -1:
            sch_id = ConfigSchema.insert(
                digest=sch_digest, schema_data=sch_data
            ).execute()
        return sch_id



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import os
from contextlib import contextmanager

from karabo.middlelayer import Hash
from karabo.middlelayer.tests.eventloop import DeviceTest
from karabo.middlelayer_devices.project_manager import ProjectManager
from karabo.project_db.exist_db.database import ProjectDatabase
from karabo.project_db.exist_db.tests.util import stop_local_database
from karabo.project_db.exist_db.util import TESTDB_ADMIN_PASSWORD

from .projectdb_util import ConsumerDevice, VerificationProjectManager


class TestProjectManager(VerificationProjectManager, DeviceTest):
    _user = "admin"
    _password = TESTDB_ADMIN_PASSWORD

    def _db_init(self):
        return ProjectDatabase(self._user, self._password,
                               test_mode=True, init_db=True)

    def stop_local_database(self):
        stop_local_database()

    @classmethod
    @contextmanager
    def lifetimeManager(cls):
        host = os.getenv('KARABO_TEST_PROJECT_DB', 'localhost')
        port = int(os.getenv('KARABO_TEST_PROJECT_DB_PORT', '8181'))
        conf = Hash("_deviceId_", "projManTest")
        conf["projectDB"] = Hash("protocol", "exist_db",
                                 "exist_db", Hash("host", host,
                                                  "port", int(port)),
                                 "testMode", True)
        cls.local = ProjectManager(conf)
        cls.consume = ConsumerDevice({"_deviceId_": "consumeTest"})
        with cls.deviceManager(cls.local, cls.consume, lead=cls.local):
            yield



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from unittest import TestCase, main

from pint import DimensionalityError

from karabo.middlelayer import Int32, QuantityValue, numeric, unit


class Tests(TestCase):
    def setUp(self):
        self.ts1 = 1 * unit.m
        self.ts1.descriptor = Int32()
        self.ts1.timestamp = 100
        self.ts2 = self.ts1 / unit.m
        self.ts2.timestamp = 200
        self.nounit = QuantityValue(3, timestamp=300)

    def test_linspace(self):
        self.assertTrue(all(numeric.linspace(3, 5, 3) == [3, 4, 5]))
        self.assertTrue(all(numeric.linspace(0, 2, 3) == [0, 1, 2]))
        self.assertTrue(all(numeric.linspace(-2, 0, 3) == [-2, -1, 0]))

        self.assertTrue(all(numeric.linspace(3 * unit.m, 5 * unit.m, 3) /
                            unit.m == [3, 4, 5]))
        self.assertTrue(all(numeric.linspace(0 * unit.m, 2 * unit.m, 3) /
                            unit.m == [0, 1, 2]))
        self.assertTrue(all(numeric.linspace(-2 * unit.m, 0 * unit.m, 3) /
                            unit.m == [-2, -1, 0]))
        self.assertTrue(all(numeric.linspace(3 * unit.m / unit.m, 5, 3) ==
                            [3, 4, 5]))
        self.assertTrue(all(numeric.linspace(3, 5 * unit.m / unit.m, 3) ==
                            [3, 4, 5]))

        self.assertEqual(numeric.linspace(self.ts1, 3 * self.ts1,
                                          self.nounit).timestamp, 300)

        with self.assertRaises(DimensionalityError):
            numeric.linspace(2 * unit.kg, 3 * unit.m, 3)

    def test_cross(self):
        self.assertTrue(all(numeric.cross([1, 2, 3], [4, 5, 6]) ==
                            [-3, 6, -3]))
        self.assertTrue(all(numeric.cross([1, 2, 3] * unit.m, [4, 5, 6]) ==
                            [-3, 6, -3] * unit.m))
        self.assertTrue(all(numeric.cross([1, 2, 3], [4, 5, 6] * unit.m) ==
                            [-3, 6, -3] * unit.m))
        self.assertTrue(
            all(numeric.cross([1, 2, 3] * unit.s, [4, 5, 6] * unit.m) ==
                [-3, 6, -3] * unit.m * unit.s))
        self.assertEqual(numeric.cross([1, 2, 3] * self.ts1,
                                       [4, 5, 6] * self.ts2).timestamp, 200)

    def test_dot(self):
        self.assertEqual(numeric.dot([1, 2, 3], [4, 5, 6]), 32)
        self.assertEqual(numeric.dot([1, 2, 3] * unit.m, [4, 5, 6]),
                         32 * unit.m)
        self.assertEqual(numeric.dot([1, 2, 3], [4, 5, 6] * unit.m),
                         32 * unit.m)
        self.assertEqual(numeric.dot([1, 2, 3] * unit.s, [4, 5, 6] * unit.m),
                         32 * unit.m * unit.s)
        self.assertEqual(numeric.dot([1, 2, 3] * self.ts1,
                                     [4, 5, 6] * self.ts2).timestamp, 200)


if __name__ == "__main__":
    main()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from abc import ABC, abstractmethod
from contextlib import ContextDecorator
from time import gmtime, strftime

from lxml import etree

from .util import ProjectDBError, to_string

DATE_FORMAT = '%Y-%m-%d %H:%M:%S'


class HandleABC(ABC):
    """Database Handle ABC

    it enforces only the basic "load/save" functionality.
    Queries are left to implementation in the derived class.
    """
    @abstractmethod
    def hasCollection(self, path):
        raise NotImplementedError

    @abstractmethod
    def removeCollection(self, path):
        raise NotImplementedError

    @abstractmethod
    def createCollection(self, path):
        raise NotImplementedError

    @abstractmethod
    def hasDocument(self, path):
        raise NotImplementedError

    @abstractmethod
    def load(self, data, path):
        raise NotImplementedError

    @abstractmethod
    def getDoc(self, name):
        raise NotImplementedError

    @abstractmethod
    def getDocument(self, name):
        raise NotImplementedError


class DatabaseBase(ContextDecorator):
    root = None

    def __init__(self):
        self._dbhandle = None

    @property
    def dbhandle(self):
        return self._dbhandle

    @dbhandle.setter
    def dbhandle(self, value):
        assert isinstance(value, HandleABC)
        self._dbhandle = value

    def onEnter(self):
        """ To Be implemented by subclasses

        returns an object subclass of `HandleABC` which is set to the
        `self.dbhandle` object attribute in the `__enter__` conte
        """

    def path(self, domain, uuid):
        # XXX: Add a '_0' suffix to keep old code from wetting its pants
        return f"{self.root}/{domain}/{uuid}_0"

    def __enter__(self):
        """Obtain a database handle.

        sanity checks on the database should be implemented in this function
        """
        self.dbhandle = self.onEnter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Clean-up action for the ProjectDatabase context. As the database
        handle doesn't carry any state nothing needs to be done here.
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """

    def domain_exists(self, domain):
        """
        Checks if a given domain exists.
        :param domain: the domain to check
        :return: True if it exists, false otherwise
        """
        path = f"{self.root}/{domain}"
        return self.dbhandle.hasCollection(path)

    def add_domain(self, domain):
        """
        Adds a domain to the project database. A domain is a top-level
        collection located directly underneath the self.root collection. When
        created the following collections will be added to the domain:
        projects, scenes, macros, configs, device_servers, and resources.

        :param domain: the name of the domain to be created
        :return:None
        :raises: ProjectDBError on Handle failure,
                 or RuntimeError if domain creation failed otherwise
        """
        path = f"{self.root}/{domain}"
        success = self.dbhandle.createCollection(path)

        if not success:
            raise RuntimeError(f"Failed to create domain at {path}")

    @staticmethod
    def _make_xml_if_needed(xml_rep):
        """
        Returns an etree xml object from xml_rep
        :param xml_rep: the xml
        :return: a root node for the xml object
        :raises: ValueError if the object passed is not of type str or type
                 etree.ElementBase
        """
        if isinstance(xml_rep, etree._Element):
            return xml_rep
        if isinstance(xml_rep, bytes):
            xml_rep = xml_rep.decode('utf-8')
        if isinstance(xml_rep, str):
            try:
                return etree.fromstring(xml_rep)
            except etree.XMLSyntaxError as e:
                raise ValueError(
                    f"XML syntax error encountered while parsing!: {e}")

        raise ValueError(f"Cannot handle type {type(xml_rep)}: {xml_rep}")

    @staticmethod
    def _make_str_if_needed(xml_rep):
        """
        Returns a string representation of xml_rep
        :param xml_rep: the xml
        :return: a string representation of xml_rep
        :raises: ValueError if the object passed is not of type str or type
                 etree.ElementBase
        """
        if isinstance(xml_rep, bytes):
            xml_rep = xml_rep.decode('utf-8')
        if isinstance(xml_rep, str):
            return xml_rep
        if isinstance(xml_rep, etree._Element):
            return to_string(xml_rep)

        raise ValueError(f"Cannot handle type {type(xml_rep)}")

    def sanitize_database(self, domain):
        """Optional Method

        Migrates a DB to the latest sane configuration"""

    def get_configurations_from_device_name(self, domain, instance_id):
        """Returns a list of configurations for a given device

        To be implemented in the derived class

        :param domain: DB domain
        :param instance_id: instance id of the device
        :return: a list of dicts:
            [{"configid": uuid of the configuration,
              "instanceid": device instance uuid in the DB},
              ...
            ]
        """
        raise NotImplementedError

    def get_configurations_from_device_name_part(self, domain, device_id_part):
        """
        Returns a list of configurations for a given device.
        :param domain: DB domain
        :param device_id_part: part of device name; search is case-insensitive.
        :return: a list of dicts:
            [{"config_id": uuid of the configuration,
              "device_uuid": device instance uuid in the DB,
              "device_id": device instance id},
              ...
            ]
        """
        raise NotImplementedError

    def get_projects_from_device(self, domain, uuid):
        """
        Returns the projects which contain a device instance with a given uuid

        To be implemented in the derived class

        :param domain: DB domain
        :param uuid: the uuid of the device instance from the database
        :return: a set containing project names
        """
        raise NotImplementedError

    def get_projects_data_from_device(self, domain, uuid):
        raise NotImplementedError

    def get_projects_with_conf(self, domain, device_id):
        """
        Returns a dict with projects and active configurations from a device
        name.

        :param domain: DB domain
        :param device_id: the device to return the information for.
        :return: a dict:
            {"project name": configuration uuid,
             ...}
        """
        configs = self.get_configurations_from_device_name(domain,
                                                           device_id)
        projects = dict()
        for config in configs:
            instance_id = config["instanceid"]
            for project in self.get_projects_from_device(domain,
                                                         instance_id):
                projects[project] = config["configid"]
        return projects

    def get_projects_with_device(self, domain, device_id_part):
        """
        Returns a dict with data about projects that contain active
        configurations for a given device.

        :param domain: DB domain
        :param device_id_part: part of name of devices for which project data
                               must be returned.
        :return: a list of dicts:
            [{"projectname": name of project,
              "date": last modification timestamp for the project,
              "uuid": uuid of projecti
              "devices": list of ids of prj devices with the given part}, ...]
        """
        configs = self.get_configurations_from_device_name_part(domain,
                                                                device_id_part)
        projects = []
        for config in configs:
            device_uuid = config["device_uuid"]
            device_id = config["device_id"]
            for prj in self.get_projects_data_from_device(domain,
                                                          device_uuid):
                prj_in_list = next((p for p in projects
                                    if p["uuid"] == prj["uuid"]), None)
                if prj_in_list:
                    # The project is already in the resulting list due to
                    # another device_id that matched the name part; add the
                    # device_id to the 'devices' attribute of the project.
                    prj_in_list["devices"].append(device_id)
                else:
                    prj["devices"] = [device_id]
                    projects.append(prj)

        return projects

    def save_item(self, domain, uuid, item_xml, overwrite=False):
        """
        Saves a item xml file into the domain. It will
        create a new entry if the item does not exist yet, or create a new
        version of the item if it does exist. In case of a versioning
        conflict the update will fail and the most current version of the
        item file is returned.

        The root node of the xml should contain a `item_type` entry identifying
        the type of the item as one of the following:

        'projects', 'scenes', 'macros', 'device_configs', 'device_servers'

        If domain does not exist it is created given a user has appropriate
        access rights.

        :param domain: the domain under which this item is to be stored
        :param uuid: the item's uuid
        :param item_xml: the xml containing the item information
        :param overwrite: defaults to False. If set to True versioning
                          information is removed prior to database injection,
                          allowing to overwrite in case of versioning
                          conflicts.

        :return: (True, dict) if successful, (false, dict) with dict
                 representing the item's versioning information. In case of
                 version conflict this will contain the information of the
                 *newest existing* entry in the database. If saving was
                 successful (no conflict), it will contain the updated
                 versioning information after saving the item.


        :raises: ProjectDBError on Handle Failure.
            RuntimeError if item saving failed otherwise.
            AttributeError if a non-supported type is passed
        """

        # create domain if necessary
        if not self.domain_exists(domain):
            self.add_domain(domain)

        # Extract some information
        try:
            # NOTE: The client might send us garbage
            item_tree = self._make_xml_if_needed(item_xml)
        except ValueError:
            msg = f'XML parse error for item "{uuid}"'
            raise ProjectDBError(msg)

        if 'user' not in item_tree.attrib:
            item_tree.attrib['user'] = 'Karabo User'
        if not item_tree.attrib.get('date'):
            item_tree.attrib['date'] = strftime(DATE_FORMAT, gmtime())
        else:
            modified, reason = self._check_for_modification(
                domain, uuid, item_tree.attrib['date'])
            if modified:
                message = "The <b>{}</b> item <b>{}</b> could not be saved: " \
                          "{}".format(item_tree.attrib.get('item_type', ''),
                                      item_tree.attrib.get('simple_name', ''),
                                      reason)
                raise ProjectDBError(message)
            # Update time stamp
            item_tree.attrib['date'] = strftime(DATE_FORMAT, gmtime())

        # XXX: Add a revision/alias to keep old code from blowing up
        item_tree.attrib['revision'] = '0'
        item_tree.attrib['alias'] = 'default'

        item_xml = self._make_str_if_needed(item_tree)
        path = self.path(domain, uuid)

        if self.dbhandle.hasDocument(path) and not overwrite:
            raise ProjectDBError("Versioning conflict. Document exists!")
        # NOTE: The underlying HTTP code needs bytes here...
        success = self.dbhandle.load(item_xml.encode('utf8'), path)
        if not success:
            raise ProjectDBError("Saving item failed!")

        meta = {}
        meta['domain'] = domain
        meta['uuid'] = uuid
        meta['date'] = item_tree.attrib['date']
        return meta



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from unittest import TestCase

from karabo.middlelayer import Configurable, Node

from ..bases import DatabaseBase
from ..util import DEFAULT_PROTOCOL, get_node, get_project


class TestPluginLoading(TestCase):
    def test_node(self):
        node = get_node()
        self.assertTrue(isinstance(node, Node))

    def test_project_default(self):
        class Manager(Configurable):
            node = get_node()

        manager = Manager()
        project = get_project(manager.node)
        self.assertTrue(isinstance(project, DatabaseBase))
        self.assertEqual(manager.node.protocol, DEFAULT_PROTOCOL)
        self.assertFalse(manager.node.testMode)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import base64
import hashlib
from datetime import datetime

from karabo.native import Hash, decodeBinary, encodeBinary

CONFIG_DB_DEVICE_ID = "deviceId"
CONFIG_DB_NAME = "name"
CONFIG_DB_DATA = "config"
CONFIG_DB_DESCRIPTION = "description"
CONFIG_DB_PRIORITY = "priority"
CONFIG_DB_SCHEMA = "schema"
CONFIG_DB_MIN_TIMEPOINT = "min_timepoint"
CONFIG_DB_MAX_TIMEPOINT = "max_timepoint"
CONFIG_DB_DIFF_TIMEPOINT = "diff_timepoint"
CONFIG_DB_TIMEPOINT = "timepoint"
CONFIG_DB_USER = "user"
CONFIG_DB_OVERWRITABLE = "overwritable"
CONFIG_DB_CONFIG_SET_ID = "config_set_id"

ISO8601_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'


class ConfigurationDBError(Exception):
    pass


def hashToBase64Bin(hash):
    """Converts a hash to its binary serialized form encoded in Base64.

    :param hash: the hash to be converted.
    :returns: a string with the base64 encoding of the binary serialized form
        of the hash.
    """
    binHash = encodeBinary(hash)
    b64Bin = base64.b64encode(binHash).decode('utf-8')
    return b64Bin


def hashFromBase64Bin(encodedHash):
    """Retrieves a hash from its binary serialized form encoded in Base64.

    :param encodedHash: the encoded hash.
    :returns: the decoded hash.
    """
    buf = base64.b64decode(encodedHash)
    hash = decodeBinary(buf)
    return hash


def schemaToBase64Bin(sch):
    """Converts a schema to its binary serialized form encoded in Base64.

    :param sch: the schema to be converted.
    :returns: a string with the base64 encoding of the binary serialized form
        of the schema.
    """
    # TODO: Fix the schema serialization in the middlelayer-api and get rid of
    #       the "hack" of piggy-backing on the hash serialization.
    hashWithSch = Hash('sch', sch)
    binSch = encodeBinary(hashWithSch)
    b64Bin = base64.b64encode(binSch).decode('utf-8')
    return b64Bin


def schemaFromBase64Bin(encodedSch):
    """Retrieves a schema from its binary serialized form encoded in Base64.

    :param encodedSch: the encoded schema.
    :returns: the decoded schema.
    """
    buf = base64.b64decode(encodedSch)
    hashWithSch = decodeBinary(buf)
    sch = hashWithSch['sch']
    return sch


def create_config_set_id(deviceIds):
    """Generates a setId that is unique for a given set of devices.

    :param deviceIds: list of deviceIds whose setId should be generated.

    :return: a sha1 hash that is unique for a given set of devices.
    """
    assert type(deviceIds) is list
    uniques = sorted({devId.upper() for devId in deviceIds})
    ids_str = '::'.join(uniques)
    return hashlib.sha1(ids_str.encode('utf-8')).hexdigest()


def datetime_from_string(str_date):
    """Obtains the datetime object equivalent to a string that is assumed to
    be in either ISO8601 or 'YYYY-MM-DD HH:mm:SS.Micro' formats.

    An exception will be thrown if the string isn't in any of the expected
    formats.
    """
    if str_date.find('T') >= 0:
        return datetime.strptime(str_date, ISO8601_FORMAT)
    else:
        return datetime.strptime(str_date, '%Y-%m-%d %H:%M:%S.%f')



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa

CREATE_CONFIG_SET_TABLE = """
    CREATE TABLE IF NOT EXISTS ConfigSet (
      id            INTEGER      PRIMARY KEY AUTOINCREMENT,
      device_set_id STRING(40)   NOT NULL,
      config_name   STRING(80)   NOT NULL,
      user          STRING(64)   DEFAULT "." NOT NULL,
      priority      INTEGER      DEFAULT 1 CHECK(priority > 0 AND priority < 4),
      description   TEXT,
      overwritable  BOOL         DEFAULT FALSE NOT NULL,

      UNIQUE(device_set_id, config_name)
    );"""

CREATE_CONFIG_TABLE = """
    CREATE TABLE IF NOT EXISTS DeviceConfig (
      id            INTEGER      PRIMARY KEY AUTOINCREMENT,
      config_set_id INTEGER      NOT NULL,
      device_id     STRING(1024) NOT NULL,
      config_data   TEXT         NOT NULL,
      schema_id     INTEGER      NOT NULL,
      timestamp     DATETIME     NOT NULL DEFAULT(STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')),

      UNIQUE(config_set_id, device_id),
      FOREIGN KEY(config_set_id) REFERENCES ConfigSet(id),
      FOREIGN KEY(schema_id) REFERENCES ConfigSchema(id)
    );"""

CREATE_SCHEMA_TABLE = """
    CREATE TABLE IF NOT EXISTS ConfigSchema (
      id           INTEGER    PRIMARY KEY AUTOINCREMENT,
      digest       STRING(40) NOT NULL,
      schema_data  TEXT       NOT NULL,

      UNIQUE(digest)
    );"""

CREATE_INDEX_SCHEMA_DIGEST_TABLE = """
    CREATE INDEX IF NOT EXISTS IdxDigest ON ConfigSchema (
      digest ASC
    );"""

CREATE_INDEX_DEVICE_ID_TABLE = """
    CREATE INDEX IF NOT EXISTS IdxDeviceId ON DeviceConfig (
      device_id ASC
    );"""

CREATE_INDEX_PRIORITY_TABLE = """
    CREATE INDEX IF NOT EXISTS IdxPriority ON ConfigSet (
      priority ASC
    );"""

CREATE_INDEX_NAME_TABLE = """
    CREATE INDEX IF NOT EXISTS IdxName ON ConfigSet (
      config_name ASC
    );"""

CREATE_INDEX_DEVICE_SET_ID_TABLE = """
    CREATE INDEX IF NOT EXISTS IdxDeviceSetId ON ConfigSet (
      device_set_id ASC
    );"""

# --------------------------------------------------------------------------
# Functions

CMD_LIST_NAME = """
        SELECT cfg_set.config_name, cfg.timestamp, cfg_set.description,
               cfg_set.priority, cfg_set.user, cfg_set.overwritable
        FROM DeviceConfig cfg, ConfigSet cfg_set
        WHERE cfg.device_id = ?
          AND cfg.config_set_id = cfg_set.id
          AND cfg_set.config_name LIKE ?
    """

CMD_LIST_NO_NAME = """
        SELECT cfg_set.config_name, cfg.timestamp, cfg_set.description,
               cfg_set.priority, cfg_set.user, cfg_set.overwritable
        FROM DeviceConfig cfg, ConfigSet cfg_set
        WHERE cfg.device_id = ?
          AND cfg.config_set_id = cfg_set.id
    """

CMD_LIST_DEVICES_PRIORITY = """
        SELECT DISTINCT cfg.device_id
        FROM DeviceConfig cfg, ConfigSet cfg_set
        WHERE cfg_set.priority = ?
          AND cfg.config_set_id = cfg_set.id
    """


def CMD_LIST_DEVS_NO_NAME(num_of_devices):
    return f"""
        SELECT *,
        ((JULIANDAY(max_timepoint)-JULIANDAY(min_timepoint))*86400) timepoint_diff_sec
    FROM (
        SELECT COUNT(cfg.device_id) num_cfgs, cfg_set.config_name,
            cfg_set.description, cfg_set.priority, cfg_set.user, cfg_set.overwritable,
            MIN(timestamp) min_timepoint,
            MAX(timestamp) max_timepoint
        FROM DeviceConfig cfg, ConfigSet cfg_set
        WHERE cfg.device_id IN ({','.join('?'*num_of_devices)})
          AND cfg.config_set_id = cfg_set.id
        GROUP BY cfg_set.config_name, cfg_set.device_set_id
    )
    WHERE num_cfgs = {num_of_devices}
    """


CMD_GET_CONFIGURATION = """
        SELECT cfg_set.config_name, cfg.timestamp, cfg.config_data,
               sch.schema_data, cfg_set.description, cfg_set.priority,
               cfg_set.user, cfg_set.overwritable
        FROM DeviceConfig cfg, ConfigSchema sch, ConfigSet cfg_set
        WHERE cfg.schema_id = sch.id
          AND cfg.config_set_id = cfg_set.id
          AND cfg.device_id = ?
          AND cfg_set.config_name = ?
    """

CMD_GET_OVERWRITABLE_CONFIG_FOR_SET = """
        SELECT id
        FROM ConfigSet
        WHERE device_set_id = ?
          AND config_name = ?
          AND overwritable = 1
    """

CMD_GET_CONFIG_FOR_SET = """
        SELECT id
        FROM ConfigSet
        WHERE device_set_id = ?
          AND config_name = ?
    """

CMD_GET_OVERWRITABLE_CONFIGS_FOR_NAME = """
        SELECT device_set_id
          FROM ConfigSet
          WHERE config_name = ?
            AND overwritable = 1
    """

CMD_GET_CONFIGS_PER_DEVICE_SET_MIN_AGE = """
        SELECT cfg_set.id set_id, cfg_set.device_set_id set_digest,
               cfg_set.config_name, cfg.id cfg_id, cfg.device_id, cfg.timestamp
        FROM DeviceConfig cfg, ConfigSet cfg_set
        WHERE cfg.config_set_id = cfg_set.id
          AND (JULIANDAY('NOW') - JULIANDAY(cfg.timestamp))*1440 > ?
        ORDER BY set_id
"""

CMD_GET_LAST_CONFIGURATION = """
        SELECT cfg_set.config_name, cfg.timestamp, cfg.config_data,
               sch.schema_data, cfg_set.description, cfg_set.priority,
               cfg_set.user, cfg_set.overwritable
        FROM DeviceConfig cfg, ConfigSchema sch, ConfigSet cfg_set
        WHERE cfg.schema_id = sch.id
          AND cfg.config_set_id = cfg_set.id
          AND cfg.device_id = ?
          AND cfg_set.priority = ?
        ORDER BY timestamp DESC
        LIMIT 1
    """


def CMD_CHECK_NAME_USED_READONLY_CONFIG(num_of_devices):
    return f"""
        SELECT COUNT(cfg.id)
        FROM DeviceConfig cfg, ConfigSet cfg_set
        WHERE cfg.config_set_id = cfg_set.id
          AND cfg_set.config_name = ?
          AND cfg.device_id IN ({','.join('?'*num_of_devices)})
          AND cfg_set.overwritable = 0
    """


def CMD_CHECK_DEVICE_LIMIT(num_of_devices):
    return f"""
    SELECT COUNT(device_id) FROM (
        SELECT device_id, COUNT(id) num_configs
        FROM DeviceConfig
        WHERE device_id IN ({','.join('?'*num_of_devices)})
        GROUP BY device_id )
    WHERE num_configs >= ?
"""


CMD_CHECK_CONFIG_ALREADY_SAVED = """
        SELECT COUNT(id) FROM DeviceConfig
        WHERE device_id = ?
          AND timestamp = strftime('%Y-%m-%d %H:%M:%f', ?)
    """

CMD_CHECK_SCHEMA_ALREADY_SAVED = """
        SELECT id FROM ConfigSchema
        WHERE digest=?
    """

CMD_INSERT_CONFIG_SET = """
        INSERT INTO ConfigSet
        (device_set_id, config_name, user, priority, description, overwritable)
        VALUES
        (?, ?, ?, ?, ?, ?)
    """

CMD_INSERT_SCHEMA = """
        INSERT INTO ConfigSchema(digest, schema_data)
        VALUES (?, ?)
    """

CMD_INSERT_CONFIGURATION_NO_TIMESTAMP = """
        INSERT INTO DeviceConfig
        (config_set_id, device_id, config_data, schema_id)
        VALUES
        (?, ?, ?, ?)
    """

CMD_INSERT_CONFIGURATION = """
        INSERT INTO DeviceConfig
        (config_set_id, device_id, timestamp, config_data, schema_id)
        VALUES
        (?, ?, strftime('%Y-%m-%d %H:%M:%f', ?), ?, ?)
    """

CMD_UPDATE_CONFIG_SET = """
        UPDATE ConfigSet
        SET
          user = ?,
          priority = ?,
          description = ?
        WHERE id = ?
    """

CMD_UPDATE_CONFIGURATION_NO_TIMESTAMP = """
        UPDATE DeviceConfig
        SET
          config_data = ?,
          schema_id = ?,
          timestamp = STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')
        WHERE config_set_id = ?
          AND device_id = ?
    """

CMD_UPDATE_CONFIGURATION = """
        UPDATE DeviceConfig
        SET
          config_data = ?,
          schema_id = ?,
          timestamp = strftime('%Y-%m-%d %H:%M:%f', ?)
        WHERE config_set_id = ?
          AND device_id = ?
    """
